from typing import Optional
from kenna.error import InvalidRegionError

GLOBAL = "global"
CA = "ca"
EU = "eu"
US = "us"

API_ENDPOINTS_BY_REGION = {
    GLOBAL: 'https://api.kennasecurity.com',
    US: 'https://api.us.kennasecurity.com',
    EU: 'https://api.eu.kennasecurity.com',
    CA: 'https://api.ca.kennasecurity.com',
}

ALL_REGIONS = list(API_ENDPOINTS_BY_REGION.keys())
ALL_ENDPOINTS = list(API_ENDPOINTS_BY_REGION.values())

DEFAULT_REGION = GLOBAL


def get_default_endpoint() -> str:
    return get_endpoint(region=DEFAULT_REGION)


def get_endpoint(region: Optional[str] = None) -> str:
    return API_ENDPOINTS_BY_REGION[region or DEFAULT_REGION]


def validate_name(region: str) -> str:
    if region not in ALL_REGIONS:
        raise InvalidRegionError("Unsupported region: {} (supported: {})".format(region, ALL_REGIONS))
    return region



from kenna.api import Kenna

import click

import hodgepodge.click
import hodgepodge.types
import hodgepodge.time


@click.group()
@click.pass_context
def dashboard_groups(_):
    pass


@dashboard_groups.command()
@click.option('--dashboard-group-id', type=int, required=True)
@click.pass_context
def get_dashboard_groups(ctx: click.Context, dashboard_group_id: int):
    api = Kenna(**ctx.obj['kenna']['config'])
    row = api.get_dashboard_group(dashboard_group_id=dashboard_group_id)
    if row:
        hodgepodge.click.echo_as_json(row)


@dashboard_groups.command()
@click.option('--dashboard-group-ids')
@click.option('--dashboard-group-names')
@click.option('--role-ids')
@click.option('--role-names')
@click.option('--min-dashboard-group-create-time')
@click.option('--max-dashboard-group-create-time')
@click.option('--min-dashboard-group-last-update-time')
@click.option('--max-dashboard-group-last-update-time')
@click.option('--limit', type=int)
@click.pass_context
def get_dashboard_groups(
        ctx: click.Context,
        dashboard_group_ids: str,
        dashboard_group_names: str,
        role_ids: str,
        role_names: str,
        min_dashboard_group_create_time: str,
        max_dashboard_group_create_time: str,
        min_dashboard_group_last_update_time: str,
        max_dashboard_group_last_update_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    for row in api.iter_dashboard_groups(
        dashboard_group_ids=dashboard_group_ids,
        dashboard_group_names=dashboard_group_names,
        role_ids=role_ids,
        role_names=role_names,
        min_dashboard_group_create_time=hodgepodge.time.to_datetime(min_dashboard_group_create_time),
        max_dashboard_group_create_time=hodgepodge.time.to_datetime(max_dashboard_group_create_time),
        min_dashboard_group_last_update_time=hodgepodge.time.to_datetime(min_dashboard_group_last_update_time),
        max_dashboard_group_last_update_time=hodgepodge.time.to_datetime(max_dashboard_group_last_update_time),
        limit=limit,
    ):
        hodgepodge.click.echo_as_json(row)


@dashboard_groups.command()
@click.option('--dashboard-group-ids')
@click.option('--dashboard-group-names')
@click.option('--role-ids')
@click.option('--role-names')
@click.option('--min-dashboard-group-create-time')
@click.option('--max-dashboard-group-create-time')
@click.option('--min-dashboard-group-last-update-time')
@click.option('--max-dashboard-group-last-update-time')
@click.option('--limit', type=int)
@click.pass_context
def count_dashboard_groups(
        ctx: click.Context,
        dashboard_group_ids: str,
        dashboard_group_names: str,
        role_ids: str,
        role_names: str,
        min_dashboard_group_create_time: str,
        max_dashboard_group_create_time: str,
        min_dashboard_group_last_update_time: str,
        max_dashboard_group_last_update_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    groups = api.iter_dashboard_groups(
        dashboard_group_ids=dashboard_group_ids,
        dashboard_group_names=dashboard_group_names,
        role_ids=role_ids,
        role_names=role_names,
        min_dashboard_group_create_time=hodgepodge.time.to_datetime(min_dashboard_group_create_time),
        max_dashboard_group_create_time=hodgepodge.time.to_datetime(max_dashboard_group_create_time),
        min_dashboard_group_last_update_time=hodgepodge.time.to_datetime(min_dashboard_group_last_update_time),
        max_dashboard_group_last_update_time=hodgepodge.time.to_datetime(max_dashboard_group_last_update_time),
        limit=limit,
    )
    count = sum(1 for _ in groups)
    click.echo(count)



from kenna.api import Kenna

import hodgepodge.time
import hodgepodge.click
import click


@click.group()
@click.pass_context
def fixes(_):
    pass


@fixes.command()
@click.option('--fix-id', type=int, required=True)
@click.pass_context
def get_fix(ctx: click.Context, fix_id: int):
    api = Kenna(**ctx.obj['kenna']['config'])
    row = api.get_fix(fix_id=fix_id)
    if row:
        hodgepodge.click.echo_as_json(row)


@fixes.command()
@click.option('--fix-ids')
@click.option('--fix-names')
@click.option('--fix-vendors')
@click.option('--cve-ids')
@click.option('--asset-ids')
@click.option('--asset-hostnames')
@click.option('--asset-ip-addresses')
@click.option('--asset-mac-addresses')
@click.option('--asset-ip-addresses')
@click.option('--asset-mac-addresses')
@click.option('--asset-group-ids')
@click.option('--asset-group-names')
@click.option('--asset-tags')
@click.option('--min-fix-create-time')
@click.option('--max-fix-create-time')
@click.option('--min-fix-last-update-time')
@click.option('--max-fix-last-update-time')
@click.option('--limit', type=int)
@click.pass_context
def get_fixes(
        ctx: click.Context,
        fix_ids: str,
        fix_names: str,
        fix_vendors: str,
        cve_ids: str,
        asset_ids: str,
        asset_hostnames: str,
        asset_ip_addresses: str,
        asset_mac_addresses: str,
        asset_group_ids: str,
        asset_group_names: str,
        asset_tags: str,
        min_fix_create_time: str,
        max_fix_create_time: str,
        min_fix_last_update_time: str,
        max_fix_last_update_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    for row in api.iter_fixes(
        fix_ids=hodgepodge.click.str_to_list_of_int(fix_ids),
        fix_names=hodgepodge.click.str_to_list_of_str(fix_names),
        fix_vendors=hodgepodge.click.str_to_list_of_str(fix_vendors),
        cve_ids=hodgepodge.click.str_to_list_of_str(cve_ids),
        asset_ids=hodgepodge.click.str_to_list_of_int(asset_ids),
        asset_hostnames=hodgepodge.click.str_to_list_of_str(asset_hostnames),
        asset_ip_addresses=hodgepodge.click.str_to_list_of_str(asset_ip_addresses),
        asset_mac_addresses=hodgepodge.click.str_to_list_of_str(asset_mac_addresses),
        asset_group_ids=hodgepodge.click.str_to_list_of_int(asset_group_ids),
        asset_group_names=hodgepodge.click.str_to_list_of_str(asset_group_names),
        asset_tags=hodgepodge.click.str_to_list_of_str(asset_tags),
        min_fix_create_time=hodgepodge.time.to_datetime(min_fix_create_time),
        max_fix_create_time=hodgepodge.time.to_datetime(max_fix_create_time),
        min_fix_last_update_time=hodgepodge.time.to_datetime(min_fix_last_update_time),
        max_fix_last_update_time=hodgepodge.time.to_datetime(max_fix_last_update_time),
        limit=limit,
    ):
        hodgepodge.click.echo_as_json(row)


@fixes.command()
@click.option('--fix-ids')
@click.option('--fix-names')
@click.option('--fix-vendors')
@click.option('--cve-ids')
@click.option('--asset-ids')
@click.option('--asset-hostnames')
@click.option('--asset-ip-addresses')
@click.option('--asset-mac-addresses')
@click.option('--asset-ip-addresses')
@click.option('--asset-mac-addresses')
@click.option('--asset-group-ids')
@click.option('--asset-group-names')
@click.option('--asset-tags')
@click.option('--min-fix-create-time')
@click.option('--max-fix-create-time')
@click.option('--min-fix-last-update-time')
@click.option('--max-fix-last-update-time')
@click.option('--limit', type=int)
@click.pass_context
def count_fixes(
        ctx: click.Context,
        fix_ids: str,
        fix_names: str,
        fix_vendors: str,
        cve_ids: str,
        asset_ids: str,
        asset_hostnames: str,
        asset_ip_addresses: str,
        asset_mac_addresses: str,
        asset_group_ids: str,
        asset_group_names: str,
        asset_tags: str,
        min_fix_create_time: str,
        max_fix_create_time: str,
        min_fix_last_update_time: str,
        max_fix_last_update_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    rows = api.iter_fixes(
        fix_ids=hodgepodge.click.str_to_list_of_int(fix_ids),
        fix_names=hodgepodge.click.str_to_list_of_str(fix_names),
        fix_vendors=hodgepodge.click.str_to_list_of_str(fix_vendors),
        cve_ids=hodgepodge.click.str_to_list_of_str(cve_ids),
        asset_ids=hodgepodge.click.str_to_list_of_int(asset_ids),
        asset_hostnames=hodgepodge.click.str_to_list_of_str(asset_hostnames),
        asset_ip_addresses=hodgepodge.click.str_to_list_of_str(asset_ip_addresses),
        asset_mac_addresses=hodgepodge.click.str_to_list_of_str(asset_mac_addresses),
        asset_group_ids=hodgepodge.click.str_to_list_of_int(asset_group_ids),
        asset_group_names=hodgepodge.click.str_to_list_of_str(asset_group_names),
        asset_tags=hodgepodge.click.str_to_list_of_str(asset_tags),
        min_fix_create_time=hodgepodge.time.to_datetime(min_fix_create_time),
        max_fix_create_time=hodgepodge.time.to_datetime(max_fix_create_time),
        min_fix_last_update_time=hodgepodge.time.to_datetime(min_fix_last_update_time),
        max_fix_last_update_time=hodgepodge.time.to_datetime(max_fix_last_update_time),
        limit=limit,
    )
    count = sum(1 for _ in rows)
    click.echo(count)



import collections
from typing import Optional, Iterator, Iterable, List, Set, Dict
from hodgepodge.requests import DEFAULT_MAX_RETRIES_ON_REDIRECT, DEFAULT_MAX_RETRIES_ON_CONNECTION_ERRORS, \
    DEFAULT_MAX_RETRIES_ON_READ_ERRORS, DEFAULT_BACKOFF_FACTOR
from kenna import DEFAULT_API_KEY, DEFAULT_REGION
from kenna.constants import APPLICATIONS, ASSETS, CONNECTORS, CONNECTOR_RUNS, \
    DASHBOARD_GROUPS, FIXES, VULNERABILITIES, USERS, ROLES, ASSET_GROUPS, COLLECTION_NAMES, OBJECT_TYPES

import datetime
import hodgepodge.archives
import hodgepodge.compression
import hodgepodge.files
import hodgepodge.hashing
import hodgepodge.network
import hodgepodge.numbers
import hodgepodge.patterns
import hodgepodge.requests
import hodgepodge.time
import hodgepodge.types
import hodgepodge.web
import itertools
import kenna.authentication
import kenna.region
import logging
import requests
import urllib.parse


logging.basicConfig(level=logging.INFO)

logger = logging.getLogger(__name__)


class Kenna:
    def __init__(
            self,
            api_key: str = DEFAULT_API_KEY,
            region: str = DEFAULT_REGION,
            max_retries_on_connection_errors: int = DEFAULT_MAX_RETRIES_ON_CONNECTION_ERRORS,
            max_retries_on_read_errors: int = DEFAULT_MAX_RETRIES_ON_READ_ERRORS,
            max_retries_on_redirects: int = DEFAULT_MAX_RETRIES_ON_REDIRECT,
            backoff_factor: float = DEFAULT_BACKOFF_FACTOR):

        self.region = region = kenna.region.validate_name(region)
        self.url = kenna.region.get_endpoint(region)

        self.session = requests.Session()
        self.session.headers['X-Risk-Token'] = kenna.authentication.validate_api_key(api_key)

        auto_retry_policy = hodgepodge.requests.get_automatic_retry_policy(
            max_retries_on_connection_errors=max_retries_on_connection_errors,
            max_retries_on_read_errors=max_retries_on_read_errors,
            max_retries_on_redirects=max_retries_on_redirects,
            backoff_factor=backoff_factor,
        )
        hodgepodge.requests.attach_http_request_policies_to_session(session=self.session, policies=[auto_retry_policy])

    def _get_object(self, url: str) -> Optional[dict]:
        response = self.session.get(url)
        if response.status_code == 404:
            return None

        response.raise_for_status()
        data = response.json()
        if len(data.keys()) == 1:
            _, v = data.popitem()
            return v
        else:
            for k, v in data.items():
                if k in OBJECT_TYPES:
                    return v
            else:
                return data

    def _get_response(self, url: str) -> Optional[dict]:
        response = self.session.get(url)
        response.raise_for_status()
        return response.json()

    def _iter_objects(self, url: str, params: Optional[dict] = None):
        params = params or {}

        response = self.session.get(url=url, params=params)
        response.raise_for_status()

        #: Read the first page of the response.
        data = response.json()
        if isinstance(data, list):
            for row in data:
                yield row

        elif isinstance(data, dict):
            for key, values in data.items():
                if key in COLLECTION_NAMES:
                    for value in values:
                        yield value

            #: Read any remaining pages (note that page numbers are 1-indexed).
            meta = data.get('meta')
            if meta:
                for page_number in range(2, meta['pages'] + 1):
                    params['page'] = page_number

                    response = self.session.get(url=url, params=params)
                    response.raise_for_status()

                    for (key, values) in response.json().items():
                        if key == 'meta':
                            continue

                        for value in values:
                            yield value
        else:
            raise TypeError("Received {} - expected one of {}".format(type(data).__name__, [t.__name__ for t in [dict, list]]))

    def get_application(self, application_id: int) -> Optional[dict]:
        url = urllib.parse.urljoin(self.url, '{}/{}'.format(APPLICATIONS, application_id))
        return self._get_object(url)

    def get_applications(
            self,
            application_ids: Optional[Iterable[int]] = None,
            application_names: Optional[Iterable[str]] = None,
            application_owners: Optional[Iterable[str]] = None,
            application_teams: Optional[Iterable[str]] = None,
            application_business_units: Optional[Iterable[str]] = None,
            min_application_risk_meter_score: Optional[int] = None,
            max_application_risk_meter_score: Optional[int] = None,
            min_asset_count: Optional[int] = None,
            max_asset_count: Optional[int] = None,
            min_vulnerability_count: Optional[int] = None,
            max_vulnerability_count: Optional[int] = None,
            limit: Optional[int] = None) -> List[dict]:

        return list(self.iter_applications(
            application_ids=application_ids,
            application_names=application_names,
            application_owners=application_owners,
            application_teams=application_teams,
            application_business_units=application_business_units,
            min_application_risk_meter_score=min_application_risk_meter_score,
            max_application_risk_meter_score=max_application_risk_meter_score,
            min_asset_count=min_asset_count,
            max_asset_count=max_asset_count,
            min_vulnerability_count=min_vulnerability_count,
            max_vulnerability_count=max_vulnerability_count,
            limit=limit,
        ))

    def iter_applications(
            self,
            application_ids: Optional[Iterable[int]] = None,
            application_names: Optional[Iterable[str]] = None,
            application_owners: Optional[Iterable[str]] = None,
            application_teams: Optional[Iterable[str]] = None,
            application_business_units: Optional[Iterable[str]] = None,
            min_application_risk_meter_score: Optional[int] = None,
            max_application_risk_meter_score: Optional[int] = None,
            min_asset_count: Optional[int] = None,
            max_asset_count: Optional[int] = None,
            min_vulnerability_count: Optional[int] = None,
            max_vulnerability_count: Optional[int] = None,
            limit: Optional[int] = None) -> Iterator[dict]:

        i = 0
        url = urllib.parse.urljoin(self.url, APPLICATIONS)
        for app in self._iter_objects(url=url):

            #: Filter applications by ID.
            if application_ids and app['id'] not in application_ids:
                continue

            #: Filter applications by name.
            if application_names and not hodgepodge.patterns.str_matches_glob(app['name'], application_names):
                continue

            #: Filter applications by by owner.
            if application_owners and not hodgepodge.patterns.str_matches_glob(app['owner'], application_owners):
                continue

            #: Filter applications by team name.
            if application_teams and not hodgepodge.patterns.str_matches_glob(app['team_name'], application_teams):
                continue

            #: Filter application by business unit.
            if application_business_units and not hodgepodge.patterns.str_matches_glob(app['business_units'], application_business_units):
                continue

            #: Filter applications by meter score.
            if (min_application_risk_meter_score is not None or max_application_risk_meter_score is not None) and \
                    not hodgepodge.numbers.is_within_range(app['risk_meter_score'], min_application_risk_meter_score, max_application_risk_meter_score):
                continue

            #: Filter applications by vulnerability count.
            if min_vulnerability_count is not None or max_vulnerability_count is not None:
                if not hodgepodge.numbers.is_within_range(
                    value=app['total_vulnerability_count'],
                    minimum=min_vulnerability_count,
                    maximum=max_vulnerability_count,
                ):
                    continue

            #: Filter applications by asset count.
            if min_asset_count is not None or max_asset_count is not None:
                if not hodgepodge.numbers.is_within_range(app['asset_count'], min_asset_count, max_asset_count):
                    continue

            yield app

            if limit:
                i += 1
                if i >= limit:
                    return

    def get_asset(self, asset_id: int) -> Optional[dict]:
        url = urllib.parse.urljoin(self.url, '{}/{}'.format(ASSETS, asset_id))
        return self._get_object(url)

    def get_assets(
            self,
            asset_ids: Optional[Iterable[int]] = None,
            asset_group_ids: Optional[Iterable[int]] = None,
            asset_group_names: Optional[Iterable[str]] = None,
            asset_tags: Optional[Iterable[str]] = None,
            asset_hostnames: Optional[Iterable[str]] = None,
            asset_ip_addresses: Optional[Iterable[str]] = None,
            asset_mac_addresses: Optional[Iterable[str]] = None,
            min_asset_risk_meter_score: Optional[int] = None,
            max_asset_risk_meter_score: Optional[int] = None,
            min_asset_first_seen_time: Optional[datetime.datetime] = None,
            max_asset_first_seen_time: Optional[datetime.datetime] = None,
            min_asset_last_seen_time: Optional[datetime.datetime] = None,
            max_asset_last_seen_time: Optional[datetime.datetime] = None,
            min_asset_last_boot_time: Optional[datetime.datetime] = None,
            max_asset_last_boot_time: Optional[datetime.datetime] = None,
            limit: Optional[int] = None) -> List[dict]:

        return list(self.iter_assets(
            asset_ids=asset_ids,
            asset_group_ids=asset_group_ids,
            asset_group_names=asset_group_names,
            asset_tags=asset_tags,
            asset_hostnames=asset_hostnames,
            asset_ip_addresses=asset_ip_addresses,
            asset_mac_addresses=asset_mac_addresses,
            min_asset_risk_meter_score=min_asset_risk_meter_score,
            max_asset_risk_meter_score=max_asset_risk_meter_score,
            min_asset_first_seen_time=min_asset_first_seen_time,
            max_asset_first_seen_time=max_asset_first_seen_time,
            min_asset_last_seen_time=min_asset_last_seen_time,
            max_asset_last_seen_time=max_asset_last_seen_time,
            min_asset_last_boot_time=min_asset_last_boot_time,
            max_asset_last_boot_time=max_asset_last_boot_time,
            limit=limit,
        ))

    def iter_assets(
            self,
            asset_ids: Optional[Iterable[int]] = None,
            asset_group_ids: Optional[Iterable[int]] = None,
            asset_group_names: Optional[Iterable[str]] = None,
            asset_tags: Optional[Iterable[str]] = None,
            asset_hostnames: Optional[Iterable[str]] = None,
            asset_ip_addresses: Optional[Iterable[str]] = None,
            asset_mac_addresses: Optional[Iterable[str]] = None,
            min_asset_risk_meter_score: Optional[int] = None,
            max_asset_risk_meter_score: Optional[int] = None,
            min_asset_first_seen_time: Optional[datetime.datetime] = None,
            max_asset_first_seen_time: Optional[datetime.datetime] = None,
            min_asset_last_seen_time: Optional[datetime.datetime] = None,
            max_asset_last_seen_time: Optional[datetime.datetime] = None,
            min_asset_last_boot_time: Optional[datetime.datetime] = None,
            max_asset_last_boot_time: Optional[datetime.datetime] = None,
            limit: Optional[int] = None) -> Iterator[dict]:

        i = 0
        url = urllib.parse.urljoin(self.url, ASSETS)
        for asset in self._iter_objects(url=url):

            #: Filter applications by ID.
            if asset_ids and asset['id'] not in asset_ids:
                continue

            #: Filter by asset group ID.
            if asset_group_ids:
                a = set(asset_group_ids)
                b = {g['id'] for g in asset['asset_groups']}
                if not a & b:
                    continue

            #: Filter by asset group name.
            if asset_group_names and not hodgepodge.patterns.str_matches_glob(
                values={g['name'] for g in asset['asset_groups']},
                patterns=asset_group_names,
            ):
                continue

            #: Filter by create time.
            if (min_asset_first_seen_time or max_asset_first_seen_time) and \
                    not hodgepodge.time.is_within_range(asset['created_at'], min_asset_first_seen_time, max_asset_first_seen_time):
                continue

            #: Filter by last seen time.
            if (min_asset_last_seen_time or max_asset_last_seen_time) and \
                    not hodgepodge.time.is_within_range(asset['last_seen_time'], min_asset_last_seen_time, max_asset_last_seen_time):
                continue

            #: Filter by last boot time.
            if (min_asset_last_boot_time or max_asset_last_boot_time) and \
                    not hodgepodge.time.is_within_range(asset['last_boot_time'], min_asset_last_boot_time, max_asset_last_boot_time):
                continue

            #: Filter by hostname.
            if asset_hostnames and not hodgepodge.patterns.str_matches_glob(
                values=[hostname for hostname in (asset['hostname'], asset['fqdn']) if hostname],
                patterns=asset_hostnames,
            ):
                continue

            #: Filter by IP address.
            if asset_ip_addresses and not hodgepodge.patterns.str_matches_glob(
                values=[ip for ip in [asset['ip_address'], asset['ipv6']] if ip],
                patterns=asset_ip_addresses,
            ):
                continue

            #: Filter by MAC address.
            if asset_mac_addresses:
                if not asset['mac_address'] or not hodgepodge.patterns.str_matches_glob(
                    values=hodgepodge.network.parse_mac_address(asset['mac_address']),
                    patterns=list(map(hodgepodge.network.parse_mac_address, asset_mac_addresses)),
                ):
                    continue

            #: Filter by tag.
            if asset_tags and not hodgepodge.patterns.str_matches_glob(asset['tags'], asset_tags):
                continue

            #: Filter by risk meter score.
            if (min_asset_risk_meter_score or max_asset_risk_meter_score) and \
                    not hodgepodge.numbers.is_within_range(
                        value=asset['risk_meter_score'],
                        minimum=min_asset_risk_meter_score,
                        maximum=max_asset_risk_meter_score,
                    ):
                continue

            yield asset

            if limit:
                i += 1
                if i >= limit:
                    return

    def get_asset_tags(
            self,
            asset_ids: Optional[Iterable[int]] = None,
            asset_group_ids: Optional[Iterable[int]] = None,
            asset_group_names: Optional[Iterable[str]] = None,
            asset_tags: Optional[Iterable[str]] = None,
            asset_hostnames: Optional[Iterable[str]] = None,
            asset_ip_addresses: Optional[Iterable[str]] = None,
            asset_mac_addresses: Optional[Iterable[str]] = None,
            min_asset_risk_meter_score: Optional[int] = None,
            max_asset_risk_meter_score: Optional[int] = None,
            min_asset_first_seen_time: Optional[datetime.datetime] = None,
            max_asset_first_seen_time: Optional[datetime.datetime] = None,
            min_asset_last_seen_time: Optional[datetime.datetime] = None,
            max_asset_last_seen_time: Optional[datetime.datetime] = None,
            min_asset_last_boot_time: Optional[datetime.datetime] = None,
            max_asset_last_boot_time: Optional[datetime.datetime] = None) -> Set[str]:

        return set(self.iter_asset_tags(
            asset_ids=asset_ids,
            asset_group_ids=asset_group_ids,
            asset_group_names=asset_group_names,
            asset_tags=asset_tags,
            asset_hostnames=asset_hostnames,
            asset_ip_addresses=asset_ip_addresses,
            asset_mac_addresses=asset_mac_addresses,
            min_asset_risk_meter_score=min_asset_risk_meter_score,
            max_asset_risk_meter_score=max_asset_risk_meter_score,
            min_asset_first_seen_time=min_asset_first_seen_time,
            max_asset_first_seen_time=max_asset_first_seen_time,
            min_asset_last_seen_time=min_asset_last_seen_time,
            max_asset_last_seen_time=max_asset_last_seen_time,
            min_asset_last_boot_time=min_asset_last_boot_time,
            max_asset_last_boot_time=max_asset_last_boot_time,
        ))

    def iter_asset_tags(
            self,
            asset_ids: Optional[Iterable[int]] = None,
            asset_group_ids: Optional[Iterable[int]] = None,
            asset_group_names: Optional[Iterable[str]] = None,
            asset_tags: Optional[Iterable[str]] = None,
            asset_hostnames: Optional[Iterable[str]] = None,
            asset_ip_addresses: Optional[Iterable[str]] = None,
            asset_mac_addresses: Optional[Iterable[str]] = None,
            min_asset_risk_meter_score: Optional[int] = None,
            max_asset_risk_meter_score: Optional[int] = None,
            min_asset_first_seen_time: Optional[datetime.datetime] = None,
            max_asset_first_seen_time: Optional[datetime.datetime] = None,
            min_asset_last_seen_time: Optional[datetime.datetime] = None,
            max_asset_last_seen_time: Optional[datetime.datetime] = None,
            min_asset_last_boot_time: Optional[datetime.datetime] = None,
            max_asset_last_boot_time: Optional[datetime.datetime] = None):

        tags = set()
        for asset in self.iter_assets(
            asset_ids=asset_ids,
            asset_group_ids=asset_group_ids,
            asset_group_names=asset_group_names,
            asset_tags=asset_tags,
            asset_hostnames=asset_hostnames,
            asset_ip_addresses=asset_ip_addresses,
            asset_mac_addresses=asset_mac_addresses,
            min_asset_risk_meter_score=min_asset_risk_meter_score,
            max_asset_risk_meter_score=max_asset_risk_meter_score,
            min_asset_first_seen_time=min_asset_first_seen_time,
            max_asset_first_seen_time=max_asset_first_seen_time,
            min_asset_last_seen_time=min_asset_last_seen_time,
            max_asset_last_seen_time=max_asset_last_seen_time,
            min_asset_last_boot_time=min_asset_last_boot_time,
            max_asset_last_boot_time=max_asset_last_boot_time,
        ):
            for tag in asset['tags']:
                if tag not in tags:
                    tags.add(tag)
                    yield tag

    def get_asset_group(self, asset_group_id: int) -> Optional[dict]:
        url = urllib.parse.urljoin(self.url, '{}/{}'.format(ASSET_GROUPS, asset_group_id))
        return self._get_object(url)

    def get_asset_groups(
            self,
            asset_group_ids: Optional[Iterable[int]] = None,
            asset_group_names: Optional[Iterable[str]] = None,
            asset_ids: Optional[Iterable[int]] = None,
            asset_tags: Optional[Iterable[str]] = None,
            asset_hostnames: Optional[Iterable[str]] = None,
            asset_ip_addresses: Optional[Iterable[str]] = None,
            asset_mac_addresses: Optional[Iterable[str]] = None,
            min_asset_group_create_time: Optional[datetime.datetime] = None,
            max_asset_group_create_time: Optional[datetime.datetime] = None,
            min_asset_group_last_update_time: Optional[datetime.datetime] = None,
            max_asset_group_last_update_time: Optional[datetime.datetime] = None,
            limit: Optional[int] = None) -> List[dict]:

        return list(self.iter_asset_groups(
            asset_group_ids=asset_group_ids,
            asset_group_names=asset_group_names,
            asset_ids=asset_ids,
            asset_tags=asset_tags,
            asset_hostnames=asset_hostnames,
            asset_ip_addresses=asset_ip_addresses,
            asset_mac_addresses=asset_mac_addresses,
            min_asset_group_create_time=min_asset_group_create_time,
            max_asset_group_create_time=max_asset_group_create_time,
            min_asset_group_last_update_time=min_asset_group_last_update_time,
            max_asset_group_last_update_time=max_asset_group_last_update_time,
            limit=limit,
        ))

    def iter_asset_groups(
            self,
            asset_group_ids: Optional[Iterable[int]] = None,
            asset_group_names: Optional[Iterable[str]] = None,
            asset_ids: Optional[Iterable[int]] = None,
            asset_tags: Optional[Iterable[str]] = None,
            asset_hostnames: Optional[Iterable[str]] = None,
            asset_ip_addresses: Optional[Iterable[str]] = None,
            asset_mac_addresses: Optional[Iterable[str]] = None,
            min_asset_group_create_time: Optional[datetime.datetime] = None,
            max_asset_group_create_time: Optional[datetime.datetime] = None,
            min_asset_group_last_update_time: Optional[datetime.datetime] = None,
            max_asset_group_last_update_time: Optional[datetime.datetime] = None,
            limit: Optional[int] = None) -> Iterator[dict]:

        #: Lookup assets.
        if asset_ids or asset_tags or asset_hostnames or asset_ip_addresses or asset_mac_addresses:
            assets = self.iter_assets(
                asset_ids=asset_ids,
                asset_tags=asset_tags,
                asset_hostnames=asset_hostnames,
                asset_ip_addresses=asset_ip_addresses,
                asset_mac_addresses=asset_mac_addresses,
            )
            asset_group_ids = {g['id'] for g in itertools.chain.from_iterable([a['asset_groups'] for a in assets])}
            asset_group_names = None

        #: Lookup asset groups.
        i = 0
        url = urllib.parse.urljoin(self.url, ASSET_GROUPS)
        for g in self._iter_objects(url=url):

            #: Filter applications by ID.
            if asset_group_ids and g['id'] not in asset_group_ids:
                continue

            #: Filter by asset group name.
            if asset_group_names and not hodgepodge.patterns.str_matches_glob(g['name'], asset_group_names):
                continue

            #: Filter by creation time.
            if (min_asset_group_create_time or max_asset_group_create_time) and \
                    not hodgepodge.time.is_within_range(g['created_at'], min_asset_group_create_time, max_asset_group_create_time):
                continue

            #: Filter by last update time.
            if (min_asset_group_last_update_time or max_asset_group_last_update_time) and \
                    not hodgepodge.time.is_within_range(g['updated_at'], min_asset_group_last_update_time, max_asset_group_last_update_time):
                continue

            yield g

            if limit:
                i += 1
                if i >= limit:
                    return

    def get_connector(self, connector_id: int) -> Optional[dict]:
        url = urllib.parse.urljoin(self.url, '{}/{}'.format(CONNECTORS, connector_id))
        return self._get_object(url)

    def get_connectors(
            self,
            connector_ids: Optional[Iterable[int]] = None,
            connector_names: Optional[Iterable[str]] = None,
            min_connector_run_start_time: Optional[datetime.datetime] = None,
            max_connector_run_start_time: Optional[datetime.datetime] = None,
            min_connector_run_end_time: Optional[datetime.datetime] = None,
            max_connector_run_end_time: Optional[datetime.datetime] = None,
            limit: Optional[int] = None) -> List[dict]:

        return list(self.iter_connectors(
            connector_ids=connector_ids,
            connector_names=connector_names,
            min_connector_run_start_time=min_connector_run_start_time,
            max_connector_run_start_time=max_connector_run_start_time,
            min_connector_run_end_time=min_connector_run_end_time,
            max_connector_run_end_time=max_connector_run_end_time,
            limit=limit,
        ))

    def iter_connectors(
            self,
            connector_ids: Optional[Iterable[int]] = None,
            connector_names: Optional[Iterable[str]] = None,
            min_connector_run_start_time: Optional[datetime.datetime] = None,
            max_connector_run_start_time: Optional[datetime.datetime] = None,
            min_connector_run_end_time: Optional[datetime.datetime] = None,
            max_connector_run_end_time: Optional[datetime.datetime] = None,
            limit: Optional[int] = None) -> Iterator[dict]:

        #: Lookup connector runs to determine target connector IDs given a min/max start/end time.
        if min_connector_run_start_time or max_connector_run_start_time or min_connector_run_end_time or \
                max_connector_run_end_time:

            runs = self.get_connector_runs_by_connector_id(
                min_connector_run_start_time=min_connector_run_start_time,
                max_connector_run_start_time=max_connector_run_start_time,
                min_connector_run_end_time=min_connector_run_end_time,
                max_connector_run_end_time=max_connector_run_end_time,
            )
            connector_ids = set(runs.keys())

        #: Lookup connector runs.
        i = 0
        url = urllib.parse.urljoin(self.url, CONNECTORS)
        for connector in self._iter_objects(url=url):

            #: Filter by ID.
            if connector_ids and connector['id'] not in connector_ids:
                continue

            #: Filter by name.
            if connector_names and not hodgepodge.patterns.str_matches_glob(connector['name'], connector_names):
                continue

            yield connector

            if limit:
                i += 1
                if i >= limit:
                    return

    def get_connector_run(self, connector_id: int, connector_run_id: int) -> Optional[dict]:
        url = urllib.parse.urljoin(self.url, '{}/{}/{}/{}'.format(CONNECTORS, connector_id, CONNECTOR_RUNS, connector_run_id))
        return self._get_object(url)

    def get_connector_runs_by_connector_id(
            self,
            connector_ids: Optional[Iterable[int]] = None,
            connector_names: Optional[Iterable[str]] = None,
            connector_run_ids: Optional[Iterable[int]] = None,
            min_connector_run_start_time: Optional[datetime.datetime] = None,
            max_connector_run_start_time: Optional[datetime.datetime] = None,
            min_connector_run_end_time: Optional[datetime.datetime] = None,
            max_connector_run_end_time: Optional[datetime.datetime] = None,
            limit: Optional[int] = None) -> Dict[int, dict]:

        connector_runs = collections.defaultdict(list)
        i = 0
        for connector in self.iter_connectors(
            connector_ids=connector_ids,
            connector_names=connector_names,
        ):
            url = urllib.parse.urljoin(self.url, '{}/{}/{}'.format(CONNECTORS, connector['id'], CONNECTOR_RUNS))
            for run in self._iter_objects(url=url):

                #: Filter by connector run ID.
                if connector_run_ids and run['id'] not in connector_run_ids:
                    continue

                #: Filter by connector run start time.
                if (min_connector_run_start_time or max_connector_run_start_time) and not \
                        hodgepodge.time.is_within_range(
                            timestamp=run['start_time'],
                            minimum=min_connector_run_start_time,
                            maximum=max_connector_run_start_time,
                        ):
                    continue

                #: Filter by connector run end time.
                if (min_connector_run_end_time or max_connector_run_end_time) and not \
                        hodgepodge.time.is_within_range(
                            timestamp=run['end_time'],
                            minimum=min_connector_run_end_time,
                            maximum=max_connector_run_end_time,
                        ):
                    continue

                connector_runs[connector['id']].append(run)

                if limit:
                    i += 1
                    if i >= limit:
                        return dict(connector_runs)
        return dict(connector_runs)

    def get_dashboard_group(self, dashboard_group_id: int) -> Optional[dict]:
        url = urllib.parse.urljoin(self.url, '{}/{}'.format(DASHBOARD_GROUPS, dashboard_group_id))
        return self._get_object(url)

    def get_dashboard_groups(
            self,
            dashboard_group_ids: Optional[Iterable[int]] = None,
            dashboard_group_names: Optional[Iterable[str]] = None,
            role_ids: Optional[Iterable[int]] = None,
            role_names: Optional[Iterable[str]] = None,
            min_dashboard_group_create_time: Optional[datetime.datetime] = None,
            max_dashboard_group_create_time: Optional[datetime.datetime] = None,
            min_dashboard_group_last_update_time: Optional[datetime.datetime] = None,
            max_dashboard_group_last_update_time: Optional[datetime.datetime] = None,
            limit: Optional[int] = None) -> List[dict]:

        return list(self.iter_dashboard_groups(
            dashboard_group_ids=dashboard_group_ids,
            dashboard_group_names=dashboard_group_names,
            role_ids=role_ids,
            role_names=role_names,
            min_dashboard_group_create_time=min_dashboard_group_create_time,
            max_dashboard_group_create_time=max_dashboard_group_create_time,
            min_dashboard_group_last_update_time=min_dashboard_group_last_update_time,
            max_dashboard_group_last_update_time=max_dashboard_group_last_update_time,
            limit=limit,
        ))

    def iter_dashboard_groups(
            self,
            dashboard_group_ids: Optional[Iterable[int]] = None,
            dashboard_group_names: Optional[Iterable[str]] = None,
            role_ids: Optional[Iterable[int]] = None,
            role_names: Optional[Iterable[str]] = None,
            min_dashboard_group_create_time: Optional[datetime.datetime] = None,
            max_dashboard_group_create_time: Optional[datetime.datetime] = None,
            min_dashboard_group_last_update_time: Optional[datetime.datetime] = None,
            max_dashboard_group_last_update_time: Optional[datetime.datetime] = None,
            limit: Optional[int] = None) -> Iterator[dict]:

        i = 0
        url = urllib.parse.urljoin(self.url, DASHBOARD_GROUPS)
        for g in self._iter_objects(url=url):

            #: Filter by ID.
            if dashboard_group_ids and g['id'] not in dashboard_group_ids:
                continue

            #: Filter by name.
            if dashboard_group_names and not hodgepodge.patterns.str_matches_glob(g['name'], dashboard_group_names):
                continue

            #: Filter by creation time.
            if (min_dashboard_group_create_time or max_dashboard_group_create_time) and \
                    not hodgepodge.time.is_within_range(g['created_at'], min_dashboard_group_create_time, max_dashboard_group_create_time):
                continue

            #: Filter by last update time.
            if (min_dashboard_group_last_update_time or max_dashboard_group_last_update_time) and \
                    not hodgepodge.time.is_within_range(g['updated_at'], min_dashboard_group_last_update_time, max_dashboard_group_last_update_time):
                continue

            #: Filter by role ID.
            if role_ids:
                a = set(role_ids)
                b = set(g['role_ids'])
                if not (a & b):
                    continue

            #: Filter by role name.
            if role_names:
                a = {role['name'] for role in g['roles']}
                b = set(role_names)
                if not hodgepodge.patterns.str_matches_glob(a, b):
                    continue

            yield g

            if limit:
                i += 1
                if i >= limit:
                    return

    def get_fix(self, fix_id: int) -> Optional[dict]:
        url = urllib.parse.urljoin(self.url, '{}/{}'.format(FIXES, fix_id))
        return self._get_object(url)

    def get_fixes(
            self,
            fix_ids: Optional[Iterable[int]] = None,
            fix_names: Optional[Iterable[str]] = None,
            fix_vendors: Optional[Iterable[str]] = None,
            cve_ids: Optional[Iterable[str]] = None,
            asset_ids: Optional[Iterable[int]] = None,
            asset_hostnames: Optional[Iterable[str]] = None,
            asset_ip_addresses: Optional[Iterable[str]] = None,
            asset_mac_addresses: Optional[Iterable[str]] = None,
            asset_group_ids: Optional[Iterable[int]] = None,
            asset_group_names: Optional[Iterable[str]] = None,
            asset_tags: Optional[Iterable[str]] = None,
            min_fix_create_time: Optional[datetime.datetime] = None,
            max_fix_create_time: Optional[datetime.datetime] = None,
            min_fix_last_update_time: Optional[datetime.datetime] = None,
            max_fix_last_update_time: Optional[datetime.datetime] = None,
            limit: Optional[int] = None) -> List[dict]:

        return list(self.iter_fixes(
            fix_ids=fix_ids,
            fix_names=fix_names,
            fix_vendors=fix_vendors,
            cve_ids=cve_ids,
            asset_ids=asset_ids,
            asset_hostnames=asset_hostnames,
            asset_ip_addresses=asset_ip_addresses,
            asset_mac_addresses=asset_mac_addresses,
            asset_group_ids=asset_group_ids,
            asset_group_names=asset_group_names,
            asset_tags=asset_tags,
            min_fix_create_time=min_fix_create_time,
            max_fix_create_time=max_fix_create_time,
            min_fix_last_update_time=min_fix_last_update_time,
            max_fix_last_update_time=max_fix_last_update_time,
            limit=limit,
        ))

    def iter_fixes(
            self,
            fix_ids: Optional[Iterable[int]] = None,
            fix_names: Optional[Iterable[str]] = None,
            fix_vendors: Optional[Iterable[str]] = None,
            cve_ids: Optional[Iterable[str]] = None,
            asset_ids: Optional[Iterable[int]] = None,
            asset_hostnames: Optional[Iterable[str]] = None,
            asset_ip_addresses: Optional[Iterable[str]] = None,
            asset_mac_addresses: Optional[Iterable[str]] = None,
            asset_group_ids: Optional[Iterable[int]] = None,
            asset_group_names: Optional[Iterable[str]] = None,
            asset_tags: Optional[Iterable[str]] = None,
            min_fix_create_time: Optional[datetime.datetime] = None,
            max_fix_create_time: Optional[datetime.datetime] = None,
            min_fix_last_update_time: Optional[datetime.datetime] = None,
            max_fix_last_update_time: Optional[datetime.datetime] = None,
            limit: Optional[int] = None) -> Iterator[dict]:

        #: Lookup assets.
        if any((asset_ids, asset_group_ids, asset_hostnames, asset_ip_addresses, asset_mac_addresses, asset_group_ids,
                asset_group_names, asset_tags)):

            assets = self.iter_assets(
                asset_ids=asset_ids,
                asset_group_ids=asset_group_ids,
                asset_group_names=asset_group_names,
                asset_tags=asset_tags,
                asset_hostnames=asset_hostnames,
                asset_ip_addresses=asset_ip_addresses,
                asset_mac_addresses=asset_mac_addresses,
            )
            asset_ids = {asset['id'] for asset in assets}

        #: Lookup fixes.
        i = 0
        url = urllib.parse.urljoin(self.url, FIXES)
        for fix in self._iter_objects(url=url):

            #: Filter by ID.
            if fix_ids and fix['id'] not in fix_ids:
                continue

            #: Filter by asset ID.
            if asset_ids:
                a = {asset['id'] for asset in fix['assets']}
                b = set(asset_ids)
                if not (a & b):
                    continue

            #: Filter by name.
            if fix_names and not hodgepodge.patterns.str_matches_glob(fix['title'], fix_names):
                continue

            #: Filter by vendor name.
            if fix_vendors and not hodgepodge.patterns.str_matches_glob(fix['vendor'], fix_vendors):
                continue

            #: Filter by CVE ID.
            if cve_ids:
                a = set(fix['cves'])
                b = set(cve_ids)
                if not (a & b):
                    continue

            #: Filter by creation time.
            if min_fix_create_time or max_fix_create_time:
                if not hodgepodge.time.is_within_range(
                    timestamp=fix['patch_publication_date'],
                    minimum=min_fix_create_time,
                    maximum=max_fix_create_time,
                ):
                    continue

            #: Filter by last update time.
            if min_fix_last_update_time or max_fix_last_update_time:
                if not hodgepodge.time.is_within_range(
                    timestamp=fix['updated_at'],
                    minimum=min_fix_last_update_time,
                    maximum=max_fix_last_update_time,
                ):
                    continue

            yield fix

            if limit:
                i += 1
                if i >= limit:
                    return

    def get_vulnerability(self, vulnerability_id: int) -> Optional[dict]:
        url = urllib.parse.urljoin(self.url, '{}/{}'.format(VULNERABILITIES, vulnerability_id))
        return self._get_object(url)

    def get_vulnerabilities(
            self,
            vulnerability_ids: Optional[Iterable[int]] = None,
            cve_ids: Optional[Iterable[str]] = None,
            fix_ids: Optional[Iterable[int]] = None,
            fix_names: Optional[Iterable[str]] = None,
            fix_vendors: Optional[Iterable[str]] = None,
            asset_ids: Optional[Iterable[int]] = None,
            asset_hostnames: Optional[Iterable[str]] = None,
            asset_ip_addresses: Optional[Iterable[str]] = None,
            asset_mac_addresses: Optional[Iterable[str]] = None,
            asset_group_ids: Optional[Iterable[int]] = None,
            asset_group_names: Optional[Iterable[str]] = None,
            asset_tags: Optional[Iterable[str]] = None,
            min_vulnerability_risk_meter_score: Optional[int] = None,
            max_vulnerability_risk_meter_score: Optional[int] = None,
            min_vulnerability_create_time: Optional[datetime.datetime] = None,
            max_vulnerability_create_time: Optional[datetime.datetime] = None,
            min_vulnerability_first_seen_time: Optional[datetime.datetime] = None,
            max_vulnerability_first_seen_time: Optional[datetime.datetime] = None,
            min_vulnerability_last_seen_time: Optional[datetime.datetime] = None,
            max_vulnerability_last_seen_time: Optional[datetime.datetime] = None,
            min_cve_publish_time: Optional[datetime.datetime] = None,
            max_cve_publish_time: Optional[datetime.datetime] = None,
            min_patch_publish_time: Optional[datetime.datetime] = None,
            max_patch_publish_time: Optional[datetime.datetime] = None,
            min_patch_due_date: Optional[datetime.date] = None,
            max_patch_due_date: Optional[datetime.date] = None,
            limit: Optional[int] = None) -> List[dict]:

        return list(self.iter_vulnerabilities(
            vulnerability_ids=vulnerability_ids,
            cve_ids=cve_ids,
            fix_ids=fix_ids,
            fix_names=fix_names,
            fix_vendors=fix_vendors,
            asset_ids=asset_ids,
            asset_hostnames=asset_hostnames,
            asset_ip_addresses=asset_ip_addresses,
            asset_mac_addresses=asset_mac_addresses,
            asset_group_ids=asset_group_ids,
            asset_group_names=asset_group_names,
            asset_tags=asset_tags,
            min_vulnerability_risk_meter_score=min_vulnerability_risk_meter_score,
            max_vulnerability_risk_meter_score=max_vulnerability_risk_meter_score,
            min_vulnerability_create_time=min_vulnerability_create_time,
            max_vulnerability_create_time=max_vulnerability_create_time,
            min_vulnerability_first_seen_time=min_vulnerability_first_seen_time,
            max_vulnerability_first_seen_time=max_vulnerability_first_seen_time,
            min_vulnerability_last_seen_time=min_vulnerability_last_seen_time,
            max_vulnerability_last_seen_time=max_vulnerability_last_seen_time,
            min_cve_publish_time=min_cve_publish_time,
            max_cve_publish_time=max_cve_publish_time,
            min_patch_publish_time=min_patch_publish_time,
            max_patch_publish_time=max_patch_publish_time,
            min_patch_due_date=min_patch_due_date,
            max_patch_due_date=max_patch_due_date,
            limit=limit,
        ))

    def iter_vulnerabilities(
            self,
            vulnerability_ids: Optional[Iterable[int]] = None,
            cve_ids: Optional[Iterable[str]] = None,
            fix_ids: Optional[Iterable[int]] = None,
            fix_names: Optional[Iterable[str]] = None,
            fix_vendors: Optional[Iterable[str]] = None,
            asset_ids: Optional[Iterable[str]] = None,
            asset_hostnames: Optional[Iterable[str]] = None,
            asset_ip_addresses: Optional[Iterable[str]] = None,
            asset_mac_addresses: Optional[Iterable[str]] = None,
            asset_group_ids: Optional[Iterable[str]] = None,
            asset_group_names: Optional[Iterable[str]] = None,
            asset_tags: Optional[Iterable[str]] = None,
            min_vulnerability_risk_meter_score: Optional[int] = None,
            max_vulnerability_risk_meter_score: Optional[int] = None,
            min_vulnerability_create_time: Optional[datetime.datetime] = None,
            max_vulnerability_create_time: Optional[datetime.datetime] = None,
            min_vulnerability_first_seen_time: Optional[datetime.datetime] = None,
            max_vulnerability_first_seen_time: Optional[datetime.datetime] = None,
            min_vulnerability_last_seen_time: Optional[datetime.datetime] = None,
            max_vulnerability_last_seen_time: Optional[datetime.datetime] = None,
            min_cve_publish_time: Optional[datetime.datetime] = None,
            max_cve_publish_time: Optional[datetime.datetime] = None,
            min_patch_publish_time: Optional[datetime.datetime] = None,
            max_patch_publish_time: Optional[datetime.datetime] = None,
            min_patch_due_date: Optional[datetime.date] = None,
            max_patch_due_date: Optional[datetime.date] = None,
            limit: Optional[int] = None) -> Iterator[dict]:

        #: Lookup assets.
        if asset_ids or asset_group_ids or asset_hostnames or asset_ip_addresses or asset_mac_addresses or \
                asset_group_ids or asset_group_names or asset_tags:

            assets = self.iter_assets(
                asset_ids=asset_ids,
                asset_group_ids=asset_group_ids,
                asset_group_names=asset_group_names,
                asset_hostnames=asset_hostnames,
                asset_ip_addresses=asset_ip_addresses,
                asset_mac_addresses=asset_mac_addresses,
                asset_tags=asset_tags,
            )
            asset_ids = {asset['id'] for asset in assets}

        #: Lookup fixes.
        if fix_ids or fix_names or fix_vendors:
            fixes = self.iter_fixes(
                fix_ids=fix_ids,
                fix_names=fix_names,
                fix_vendors=fix_vendors,
            )
            fix_ids = {fix['id'] for fix in fixes}

        #: Lookup vulnerabilities.
        i = 0
        url = urllib.parse.urljoin(self.url, '{}'.format(VULNERABILITIES))
        for vulnerability in self._iter_objects(url=url):

            #: Filter by vulnerability ID.
            if vulnerability_ids and vulnerability['id'] not in vulnerability_ids:
                continue

            #: Filter by asset ID.
            if asset_ids and vulnerability['asset_id'] not in asset_ids:
                continue

            #: Filter by fix ID.
            if fix_ids and vulnerability['fix_id'] not in fix_ids:
                continue

            #: Filter by CVE ID.
            if cve_ids and vulnerability['cve_id'] not in cve_ids:
                continue

            #: Filter by risk meter score.
            if (min_vulnerability_risk_meter_score or max_vulnerability_risk_meter_score) and not hodgepodge.numbers.is_within_range(
                value=vulnerability['risk_meter_score'],
                minimum=min_vulnerability_risk_meter_score,
                maximum=max_vulnerability_risk_meter_score,
            ):
                continue

            #: Filter by create time.
            if (min_vulnerability_create_time or max_vulnerability_create_time) and not hodgepodge.time.is_within_range(
                timestamp=vulnerability['created_at'],
                minimum=min_vulnerability_create_time,
                maximum=max_vulnerability_create_time,
            ):
                continue

            #: Filter by first seen time.
            if (min_vulnerability_first_seen_time or max_vulnerability_first_seen_time) and not hodgepodge.time.is_within_range(
                timestamp=vulnerability['first_found_on'],
                minimum=min_vulnerability_first_seen_time,
                maximum=max_vulnerability_first_seen_time,
            ):
                continue

            #: Filter by last seen time.
            if (min_vulnerability_last_seen_time or max_vulnerability_last_seen_time) and not hodgepodge.time.is_within_range(
                timestamp=vulnerability['last_seen_time'],
                minimum=min_vulnerability_last_seen_time,
                maximum=max_vulnerability_last_seen_time,
            ):
                continue

            #: Filter by CVE publish time.
            if (min_cve_publish_time or max_cve_publish_time) and not hodgepodge.time.is_within_range(
                timestamp=vulnerability['cve_published_at'],
                minimum=min_cve_publish_time,
                maximum=max_cve_publish_time,
            ):
                continue

            #: Filter by patch publish time.
            if (min_patch_publish_time or max_patch_publish_time) and not hodgepodge.time.is_within_range(
                timestamp=vulnerability['patch_published_at'],
                minimum=min_patch_publish_time,
                maximum=max_patch_publish_time,
            ):
                continue

            #: Filter by patch due date.
            if (min_patch_due_date or max_patch_due_date) and not hodgepodge.time.is_within_range(
                timestamp=vulnerability['due_date'],
                minimum=min_patch_due_date,
                maximum=max_patch_due_date,
            ):
                continue

            yield vulnerability

            if limit:
                i += 1
                if i >= limit:
                    return

    def get_user(self, user_id: int):
        url = urllib.parse.urljoin(self.url, '{}/{}'.format(USERS, user_id))
        return self._get_object(url)

    def get_users(
            self,
            user_ids: Optional[Iterable[int]] = None,
            user_names: Optional[Iterable[str]] = None,
            user_email_addresses: Optional[Iterable[str]] = None,
            role_ids: Optional[Iterable[int]] = None,
            role_names: Optional[Iterable[str]] = None,
            min_user_create_time: Optional[datetime.datetime] = None,
            max_user_create_time: Optional[datetime.datetime] = None,
            min_user_last_update_time: Optional[datetime.datetime] = None,
            max_user_last_update_time: Optional[datetime.datetime] = None,
            min_user_last_sign_in_time: Optional[datetime.datetime] = None,
            max_user_last_sign_in_time: Optional[datetime.datetime] = None,
            limit: Optional[int] = None) -> List[dict]:

        return list(self.iter_users(
            user_ids=user_ids,
            user_names=user_names,
            user_email_addresses=user_email_addresses,
            role_ids=role_ids,
            role_names=role_names,
            min_user_create_time=min_user_create_time,
            max_user_create_time=max_user_create_time,
            min_user_last_update_time=min_user_last_update_time,
            max_user_last_update_time=max_user_last_update_time,
            min_user_last_sign_in_time=min_user_last_sign_in_time,
            max_user_last_sign_in_time=max_user_last_sign_in_time,
            limit=limit,
        ))

    def iter_users(
            self,
            user_ids: Optional[Iterable[int]] = None,
            user_names: Optional[Iterable[str]] = None,
            user_email_addresses: Optional[Iterable[str]] = None,
            role_ids: Optional[Iterable[int]] = None,
            role_names: Optional[Iterable[str]] = None,
            min_user_create_time: Optional[datetime.datetime] = None,
            max_user_create_time: Optional[datetime.datetime] = None,
            min_user_last_update_time: Optional[datetime.datetime] = None,
            max_user_last_update_time: Optional[datetime.datetime] = None,
            min_user_last_sign_in_time: Optional[datetime.datetime] = None,
            max_user_last_sign_in_time: Optional[datetime.datetime] = None,
            limit: Optional[int] = None) -> Iterator[dict]:

        i = 0
        url = urllib.parse.urljoin(self.url, USERS)
        for user in self._iter_objects(url):

            #: Filter by ID.
            if user_ids and user['id'] not in user_ids:
                continue

            #: Filter by creation time.
            if (min_user_create_time or max_user_create_time) and not hodgepodge.time.is_within_range(
                timestamp=user['created_at'],
                minimum=min_user_create_time,
                maximum=max_user_create_time,
            ):
                continue

            #: Filter by last update time.
            if (min_user_last_update_time or max_user_last_update_time) and not hodgepodge.time.is_within_range(
                timestamp=user['updated_at'],
                minimum=min_user_last_update_time,
                maximum=max_user_last_update_time,
            ):
                continue

            #: Filter by last sign-in time.
            if (min_user_last_sign_in_time or max_user_last_sign_in_time) and not hodgepodge.time.is_within_range(
                timestamp=user['last_sign_in_at'],
                minimum=min_user_last_sign_in_time,
                maximum=max_user_last_sign_in_time,
            ):
                continue

            #: Filter by name.
            if user_names:
                values = {
                    user['firstname'],
                    user['lastname'],
                    user['firstname'] + ' ' + user['lastname'],
                    user['external_id'],
                    user['email'],
                }
                if not hodgepodge.patterns.str_matches_glob(values, user_names):
                    continue

            #: Filter by email address.
            if user_email_addresses and not hodgepodge.patterns.str_matches_glob(user['email'], user_email_addresses):
                continue

            #: Filter by role ID.
            if role_ids:
                a = set(role_ids)
                b = set(user['role_ids'])
                if not (a & b):
                    continue

            #: Filter by role name.
            if role_names and not hodgepodge.patterns.str_matches_glob(user['roles'], role_names):
                continue

            yield user

            if limit:
                i += 1
                if i >= limit:
                    return

    def get_role(self, role_id: int) -> Optional[dict]:
        url = urllib.parse.urljoin(self.url, '{}/{}'.format(ROLES, role_id))
        return self._get_object(url)

    def get_roles(
            self,
            role_ids: Optional[Iterable[int]] = None,
            role_names: Optional[Iterable[str]] = None,
            role_types: Optional[Iterable[str]] = None,
            role_access_levels: Optional[Iterable[str]] = None,
            role_custom_permissions: Optional[Iterable[str]] = None,
            user_ids: Optional[Iterable[int]] = None,
            user_email_addresses: Optional[Iterable[str]] = None,
            user_names: Optional[Iterable[str]] = None,
            min_role_create_time: Optional[datetime.datetime] = None,
            max_role_create_time: Optional[datetime.datetime] = None,
            min_role_last_update_time: Optional[datetime.datetime] = None,
            max_role_last_update_time: Optional[datetime.datetime] = None,
            limit: Optional[int] = None) -> List[dict]:

        return list(self.iter_roles(
            role_ids=role_ids,
            role_names=role_names,
            role_types=role_types,
            role_access_levels=role_access_levels,
            role_custom_permissions=role_custom_permissions,
            user_ids=user_ids,
            user_email_addresses=user_email_addresses,
            user_names=user_names,
            min_role_create_time=min_role_create_time,
            max_role_create_time=max_role_create_time,
            min_role_last_update_time=min_role_last_update_time,
            max_role_last_update_time=max_role_last_update_time,
            limit=limit,
        ))

    def iter_roles(
            self,
            role_ids: Optional[Iterable[int]] = None,
            role_names: Optional[Iterable[str]] = None,
            role_types: Optional[Iterable[str]] = None,
            role_access_levels: Optional[Iterable[str]] = None,
            role_custom_permissions: Optional[Iterable[str]] = None,
            user_ids: Optional[Iterable[int]] = None,
            user_email_addresses: Optional[Iterable[str]] = None,
            user_names: Optional[Iterable[str]] = None,
            min_role_create_time: Optional[datetime.datetime] = None,
            max_role_create_time: Optional[datetime.datetime] = None,
            min_role_last_update_time: Optional[datetime.datetime] = None,
            max_role_last_update_time: Optional[datetime.datetime] = None,
            limit: Optional[int] = None) -> Iterator[dict]:

        #: Lookup users.
        if user_ids or user_names:
            users = self.iter_users(
                user_ids=user_ids,
                user_email_addresses=user_email_addresses,
                user_names=user_names
            )
            role_ids = set(itertools.chain.from_iterable(user['role_ids'] for user in users))

        #: Lookup roles.
        i = 0
        url = urllib.parse.urljoin(self.url, ROLES)
        for role in self._iter_objects(url):

            #: Filter by ID.
            if role_ids and role['id'] not in role_ids:
                continue

            #: Filter by name.
            if role_names and not hodgepodge.patterns.str_matches_glob(role['name'], role_names):
                continue

            #: Filter by type.
            if role_types and not hodgepodge.patterns.str_matches_glob(role['role_type'], role_types):
                continue

            #: Filter by access level.
            if role_access_levels and not hodgepodge.patterns.str_matches_glob(role['access_level'], role_access_levels):
                continue

            #: Filter by permission name.
            if role_custom_permissions and not hodgepodge.patterns.str_matches_glob(role['custom_permissions'], role_custom_permissions):
                continue

            #: Filter by creation time.
            if (min_role_create_time or max_role_create_time) and not hodgepodge.time.is_within_range(
                timestamp=role['created_at'],
                minimum=min_role_create_time,
                maximum=max_role_create_time,
            ):
                continue

            #: Filter by last update time.
            if (min_role_last_update_time or max_role_last_update_time) and not hodgepodge.time.is_within_range(
                timestamp=role['updated_at'],
                minimum=min_role_last_update_time,
                maximum=max_role_last_update_time,
            ):
                continue

            yield role

            if limit:
                i += 1
                if i >= limit:
                    return

    def get_business_units(
            self,
            application_ids: Optional[Optional[Iterable[str]]],
            application_names: Optional[Optional[Iterable[str]]] = None) -> List[dict]:

        apps = self.iter_applications(
            application_ids=application_ids,
            application_names=application_names,
        )
        names = {app['business_units'] for app in apps if app['business_units']}
        units = [{'name': name} for name in names]
        return units

    def get_teams(
            self,
            application_ids: Optional[Optional[Iterable[str]]],
            application_names: Optional[Optional[Iterable[str]]] = None) -> List[dict]:

        apps = self.iter_applications(
            application_ids=application_ids,
            application_names=application_names,
        )
        names = {app['team_name'] for app in apps if app['team_name']}
        teams = [{'name': name} for name in names]
        return teams



MAX_PAGES_ALLOWED_BY_API = 20

APPLICATIONS = 'applications'
ASSETS = 'assets'
ASSET_GROUPS = 'asset_groups'
CONNECTORS = 'connectors'
CONNECTOR_RUNS = 'connector_runs'
DASHBOARD_GROUPS = 'dashboard_groups'
FIXES = 'fixes'
ROLES = 'roles'
USERS = 'users'
VULNERABILITIES = 'vulnerabilities'
VULNERABILITY_DEFINITIONS = 'vulnerability_definitions'

COLLECTION_NAMES = [
    APPLICATIONS,
    ASSETS,
    ASSET_GROUPS,
    CONNECTORS,
    CONNECTOR_RUNS,
    DASHBOARD_GROUPS,
    FIXES,
    ROLES,
    USERS,
    VULNERABILITIES,
    VULNERABILITY_DEFINITIONS,
]

APPLICATION = 'application'
ASSET = 'asset'
ASSET_GROUP = 'asset_group'
CONNECTOR = 'connector'
CONNECTOR_RUN = 'connector_run'
DASHBOARD_GROUP = 'dashboard_group'
FIX = 'fix'
ROLE = 'role'
USER = 'user'
VULNERABILITY = 'vulnerability'
VULNERABILITY_DEFINITION = 'vulnerability_definition'

OBJECT_TYPES = [
    APPLICATION,
    ASSET,
    ASSET_GROUP,
    CONNECTOR,
    CONNECTOR_RUN,
    DASHBOARD_GROUP,
    FIX,
    ROLE,
    USER,
    VULNERABILITY,
    VULNERABILITY_DEFINITION,
]




import click

import hodgepodge.click
import hodgepodge.types
from kenna.api import Kenna


@click.group()
@click.pass_context
def applications(_):
    pass


@applications.command()
@click.option('--application-id', type=int, required=True)
@click.pass_context
def get_application(ctx: click.Context, application_id: int):
    api = Kenna(**ctx.obj['kenna']['config'])
    row = api.get_application(application_id=application_id)
    if row:
        hodgepodge.click.echo_as_json(row)


@applications.command()
@click.option('--application-ids')
@click.option('--application-names')
@click.option('--application-owners')
@click.option('--application-teams')
@click.option('--application-business-units')
@click.option('--min-application-risk-meter-score', type=int)
@click.option('--max-application-risk-meter-score', type=int)
@click.option('--min-asset-count', type=int)
@click.option('--max-asset-count', type=int)
@click.option('--min-vulnerability-count', type=int)
@click.option('--max-vulnerability-count', type=int)
@click.option('--limit', type=int)
@click.pass_context
def get_applications(
        ctx: click.Context,
        application_ids: str,
        application_names: str,
        application_owners: str,
        application_teams: str,
        application_business_units: str,
        min_application_risk_meter_score: int,
        max_application_risk_meter_score: int,
        min_asset_count: int,
        max_asset_count: int,
        min_vulnerability_count: int,
        max_vulnerability_count: int,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    for row in api.iter_applications(
        application_ids=hodgepodge.click.str_to_list_of_int(application_ids),
        application_names=hodgepodge.click.str_to_list_of_str(application_names),
        application_owners=hodgepodge.click.str_to_list_of_str(application_owners),
        application_teams=hodgepodge.click.str_to_list_of_str(application_teams),
        application_business_units=hodgepodge.click.str_to_list_of_str(application_business_units),
        min_application_risk_meter_score=min_application_risk_meter_score,
        max_application_risk_meter_score=max_application_risk_meter_score,
        min_asset_count=min_asset_count,
        max_asset_count=max_asset_count,
        min_vulnerability_count=min_vulnerability_count,
        max_vulnerability_count=max_vulnerability_count,
        limit=limit,
    ):
        hodgepodge.click.echo_as_json(row)


@applications.command()
@click.option('--application-ids')
@click.option('--application-names')
@click.option('--application-owners')
@click.option('--application-teams')
@click.option('--application-business-units')
@click.option('--min-application-risk-meter-score', type=int)
@click.option('--max-application-risk-meter-score', type=int)
@click.option('--min-asset-count', type=int)
@click.option('--max-asset-count', type=int)
@click.option('--min-vulnerability-count', type=int)
@click.option('--max-vulnerability-count', type=int)
@click.option('--limit', type=int)
@click.pass_context
def count_applications(
        ctx: click.Context,
        application_ids: str,
        application_names: str,
        application_owners: str,
        application_teams: str,
        application_business_units: str,
        min_application_risk_meter_score: int,
        max_application_risk_meter_score: int,
        min_asset_count: int,
        max_asset_count: int,
        min_vulnerability_count: int,
        max_vulnerability_count: int,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    rows = api.iter_applications(
        application_ids=hodgepodge.click.str_to_list_of_int(application_ids),
        application_names=hodgepodge.click.str_to_list_of_str(application_names),
        application_owners=hodgepodge.click.str_to_list_of_str(application_owners),
        application_teams=hodgepodge.click.str_to_list_of_str(application_teams),
        application_business_units=hodgepodge.click.str_to_list_of_str(application_business_units),
        min_application_risk_meter_score=min_application_risk_meter_score,
        max_application_risk_meter_score=max_application_risk_meter_score,
        min_asset_count=min_asset_count,
        max_asset_count=max_asset_count,
        min_vulnerability_count=min_vulnerability_count,
        max_vulnerability_count=max_vulnerability_count,
        limit=limit,
    )
    count = sum(1 for _ in rows)
    click.echo(count)



from kenna.api import Kenna

import click

import hodgepodge.time
import hodgepodge.click
import hodgepodge.types


@click.group()
@click.pass_context
def vulnerabilities(_):
    pass


@vulnerabilities.command()
@click.option('--vulnerability-id', type=int, required=True)
@click.pass_context
def get_vulnerability(ctx: click.Context, vulnerability_id: int):
    api = Kenna(**ctx.obj['kenna']['config'])
    row = api.get_vulnerability(vulnerability_id=vulnerability_id)
    if row:
        hodgepodge.click.echo_as_json(row)


@vulnerabilities.command()
@click.option('--vulnerability-ids')
@click.option('--cve-ids')
@click.option('--fix-ids')
@click.option('--fix-names')
@click.option('--fix-vendors')
@click.option('--asset-ids')
@click.option('--asset-hostnames')
@click.option('--asset-ip-addresses')
@click.option('--asset-mac-addresses')
@click.option('--asset-group-ids')
@click.option('--asset-group-names')
@click.option('--asset-tags')
@click.option('--min-vulnerability-risk-meter-score', type=int)
@click.option('--max-vulnerability-risk-meter-score', type=int)
@click.option('--min-vulnerability-first-seen-time')
@click.option('--max-vulnerability-first-seen-time')
@click.option('--min-vulnerability-last-seen-time')
@click.option('--max-vulnerability-last-seen-time')
@click.option('--min-cve-publish-time')
@click.option('--max-cve-publish-time')
@click.option('--min-patch-publish-time')
@click.option('--max-patch-publish-time')
@click.option('--min-patch-due-date')
@click.option('--max-patch-due-date')
@click.option('--limit', type=int)
@click.pass_context
def get_vulnerabilities(
        ctx: click.Context,
        vulnerability_ids: str,
        cve_ids: str,
        fix_ids: str,
        fix_names: str,
        fix_vendors: str,
        asset_ids: str,
        asset_hostnames: str,
        asset_ip_addresses: str,
        asset_mac_addresses: str,
        asset_group_ids: str,
        asset_group_names: str,
        asset_tags: str,
        min_vulnerability_risk_meter_score: int,
        max_vulnerability_risk_meter_score: int,
        min_vulnerability_first_seen_time: str,
        max_vulnerability_first_seen_time: str,
        min_vulnerability_last_seen_time: str,
        max_vulnerability_last_seen_time: str,
        min_cve_publish_time: str,
        max_cve_publish_time: str,
        min_patch_publish_time: str,
        max_patch_publish_time: str,
        min_patch_due_date: str,
        max_patch_due_date: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    for row in api.get_vulnerabilities(
        vulnerability_ids=hodgepodge.click.str_to_list_of_int(vulnerability_ids),
        cve_ids=hodgepodge.click.str_to_list_of_str(cve_ids),
        fix_ids=hodgepodge.click.str_to_list_of_int(fix_ids),
        fix_names=hodgepodge.click.str_to_list_of_str(fix_names),
        fix_vendors=hodgepodge.click.str_to_list_of_str(fix_vendors),
        asset_ids=hodgepodge.click.str_to_list_of_int(asset_ids),
        asset_hostnames=hodgepodge.click.str_to_list_of_str(asset_hostnames),
        asset_ip_addresses=hodgepodge.click.str_to_list_of_str(asset_ip_addresses),
        asset_mac_addresses=hodgepodge.click.str_to_list_of_str(asset_mac_addresses),
        asset_group_ids=hodgepodge.click.str_to_list_of_int(asset_group_ids),
        asset_group_names=hodgepodge.click.str_to_list_of_str(asset_group_names),
        asset_tags=hodgepodge.click.str_to_list_of_str(asset_tags),
        min_vulnerability_risk_meter_score=min_vulnerability_risk_meter_score,
        max_vulnerability_risk_meter_score=max_vulnerability_risk_meter_score,
        min_vulnerability_first_seen_time=hodgepodge.time.to_datetime(min_vulnerability_first_seen_time),
        max_vulnerability_first_seen_time=hodgepodge.time.to_datetime(max_vulnerability_first_seen_time),
        min_vulnerability_last_seen_time=hodgepodge.time.to_datetime(min_vulnerability_last_seen_time),
        max_vulnerability_last_seen_time=hodgepodge.time.to_datetime(max_vulnerability_last_seen_time),
        min_cve_publish_time=hodgepodge.time.to_datetime(min_cve_publish_time),
        max_cve_publish_time=hodgepodge.time.to_datetime(max_cve_publish_time),
        min_patch_publish_time=hodgepodge.time.to_datetime(min_patch_publish_time),
        max_patch_publish_time=hodgepodge.time.to_datetime(max_patch_publish_time),
        min_patch_due_date=hodgepodge.time.to_datetime(min_patch_due_date),
        max_patch_due_date=hodgepodge.time.to_datetime(max_patch_due_date),
        limit=limit,
    ):
        hodgepodge.click.echo_as_json(row)


@vulnerabilities.command()
@click.option('--vulnerability-ids')
@click.option('--cve-ids')
@click.option('--fix-ids')
@click.option('--fix-names')
@click.option('--fix-vendors')
@click.option('--asset-ids')
@click.option('--asset-hostnames')
@click.option('--asset-ip-addresses')
@click.option('--asset-mac-addresses')
@click.option('--asset-group-ids')
@click.option('--asset-group-names')
@click.option('--asset-tags')
@click.option('--min-vulnerability-risk-meter-score', type=int)
@click.option('--max-vulnerability-risk-meter-score', type=int)
@click.option('--min-vulnerability-first-seen-time')
@click.option('--max-vulnerability-first-seen-time')
@click.option('--min-vulnerability-last-seen-time')
@click.option('--max-vulnerability-last-seen-time')
@click.option('--min-cve-publish-time')
@click.option('--max-cve-publish-time')
@click.option('--min-patch-publish-time')
@click.option('--max-patch-publish-time')
@click.option('--min-patch-due-date')
@click.option('--max-patch-due-date')
@click.option('--limit', type=int)
@click.pass_context
def get_vulnerabilities(
        ctx: click.Context,
        vulnerability_ids: str,
        cve_ids: str,
        fix_ids: str,
        fix_names: str,
        fix_vendors: str,
        asset_ids: str,
        asset_hostnames: str,
        asset_ip_addresses: str,
        asset_mac_addresses: str,
        asset_group_ids: str,
        asset_group_names: str,
        asset_tags: str,
        min_vulnerability_risk_meter_score: int,
        max_vulnerability_risk_meter_score: int,
        min_vulnerability_first_seen_time: str,
        max_vulnerability_first_seen_time: str,
        min_vulnerability_last_seen_time: str,
        max_vulnerability_last_seen_time: str,
        min_cve_publish_time: str,
        max_cve_publish_time: str,
        min_patch_publish_time: str,
        max_patch_publish_time: str,
        min_patch_due_date: str,
        max_patch_due_date: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    rows = api.get_vulnerabilities(
        vulnerability_ids=hodgepodge.click.str_to_list_of_int(vulnerability_ids),
        cve_ids=hodgepodge.click.str_to_list_of_str(cve_ids),
        fix_ids=hodgepodge.click.str_to_list_of_int(fix_ids),
        fix_names=hodgepodge.click.str_to_list_of_str(fix_names),
        fix_vendors=hodgepodge.click.str_to_list_of_str(fix_vendors),
        asset_ids=hodgepodge.click.str_to_list_of_int(asset_ids),
        asset_hostnames=hodgepodge.click.str_to_list_of_str(asset_hostnames),
        asset_ip_addresses=hodgepodge.click.str_to_list_of_str(asset_ip_addresses),
        asset_mac_addresses=hodgepodge.click.str_to_list_of_str(asset_mac_addresses),
        asset_group_ids=hodgepodge.click.str_to_list_of_int(asset_group_ids),
        asset_group_names=hodgepodge.click.str_to_list_of_str(asset_group_names),
        asset_tags=hodgepodge.click.str_to_list_of_str(asset_tags),
        min_vulnerability_risk_meter_score=min_vulnerability_risk_meter_score,
        max_vulnerability_risk_meter_score=max_vulnerability_risk_meter_score,
        min_vulnerability_first_seen_time=hodgepodge.time.to_datetime(min_vulnerability_first_seen_time),
        max_vulnerability_first_seen_time=hodgepodge.time.to_datetime(max_vulnerability_first_seen_time),
        min_vulnerability_last_seen_time=hodgepodge.time.to_datetime(min_vulnerability_last_seen_time),
        max_vulnerability_last_seen_time=hodgepodge.time.to_datetime(max_vulnerability_last_seen_time),
        min_cve_publish_time=hodgepodge.time.to_datetime(min_cve_publish_time),
        max_cve_publish_time=hodgepodge.time.to_datetime(max_cve_publish_time),
        min_patch_publish_time=hodgepodge.time.to_datetime(min_patch_publish_time),
        max_patch_publish_time=hodgepodge.time.to_datetime(max_patch_publish_time),
        min_patch_due_date=hodgepodge.time.to_datetime(min_patch_due_date),
        max_patch_due_date=hodgepodge.time.to_datetime(max_patch_due_date),
        limit=limit,
    )
    count = sum(1 for _ in rows)
    click.echo(count)



from kenna.api import Kenna

import click

import hodgepodge.time
import hodgepodge.click
import hodgepodge.types


@click.group()
@click.pass_context
def connectors(_):
    pass


@connectors.command()
@click.option('--connector-id', type=int, required=True)
@click.pass_context
def get_connector(ctx: click.Context, connector_id: int):
    api = Kenna(**ctx.obj['kenna']['config'])
    row = api.get_connector(connector_id=connector_id)
    hodgepodge.click.echo_as_json(row)


@connectors.command()
@click.option('--connector-id', type=int, required=True)
@click.option('--connector-run-id', type=int, required=True)
@click.pass_context
def get_connector_run(ctx: click.Context, connector_id: int, connector_run_id: int):
    api = Kenna(**ctx.obj['kenna']['config'])
    row = api.get_connector_run(
        connector_id=connector_id,
        connector_run_id=connector_run_id,
    )
    if row:
        hodgepodge.click.echo_as_json(row)


@connectors.command()
@click.option('--connector-ids')
@click.option('--connector-names')
@click.option('--min-connector-run-start-time')
@click.option('--max-connector-run-start-time')
@click.option('--min-connector-run-end-time')
@click.option('--max-connector-run-end-time')
@click.option('--limit', type=int)
@click.pass_context
def get_connectors(
        ctx: click.Context,
        connector_ids: str,
        connector_names: str,
        min_connector_run_start_time: str,
        max_connector_run_start_time: str,
        min_connector_run_end_time: str,
        max_connector_run_end_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    for row in api.iter_connectors(
        connector_ids=connector_ids,
        connector_names=connector_names,
        min_connector_run_start_time=hodgepodge.time.to_datetime(min_connector_run_start_time),
        max_connector_run_start_time=hodgepodge.time.to_datetime(max_connector_run_start_time),
        min_connector_run_end_time=hodgepodge.time.to_datetime(min_connector_run_end_time),
        max_connector_run_end_time=hodgepodge.time.to_datetime(max_connector_run_end_time),
        limit=limit,
    ):
        hodgepodge.click.echo_as_json(row)


@connectors.command()
@click.option('--connector-ids')
@click.option('--connector-names')
@click.option('--connector-run-ids')
@click.option('--min-connector-run-start-time')
@click.option('--max-connector-run-start-time')
@click.option('--min-connector-run-end-time')
@click.option('--max-connector-run-end-time')
@click.option('--limit', type=int)
@click.pass_context
def get_connector_runs(
        ctx: click.Context,
        connector_ids: str,
        connector_names: str,
        connector_run_ids: str,
        min_connector_run_start_time: str,
        max_connector_run_start_time: str,
        min_connector_run_end_time: str,
        max_connector_run_end_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    runs = api.get_connector_runs_by_connector_id(
        connector_ids=hodgepodge.click.str_to_list_of_int(connector_ids),
        connector_names=hodgepodge.click.str_to_list_of_str(connector_names),
        connector_run_ids=hodgepodge.click.str_to_list_of_int(connector_run_ids),
        min_connector_run_start_time=hodgepodge.time.to_datetime(min_connector_run_start_time),
        max_connector_run_start_time=hodgepodge.time.to_datetime(max_connector_run_start_time),
        min_connector_run_end_time=hodgepodge.time.to_datetime(min_connector_run_end_time),
        max_connector_run_end_time=hodgepodge.time.to_datetime(max_connector_run_end_time),
        limit=limit,
    )
    hodgepodge.click.echo_as_json(runs)


@connectors.command()
@click.option('--connector-ids')
@click.option('--connector-names')
@click.option('--min-connector-run-start-time')
@click.option('--max-connector-run-start-time')
@click.option('--min-connector-run-end-time')
@click.option('--max-connector-run-end-time')
@click.option('--limit', type=int)
@click.pass_context
def count_connectors(
        ctx: click.Context,
        connector_ids: str,
        connector_names: str,
        min_connector_run_start_time: str,
        max_connector_run_start_time: str,
        min_connector_run_end_time: str,
        max_connector_run_end_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    rows = api.iter_connectors(
        connector_ids=connector_ids,
        connector_names=connector_names,
        min_connector_run_start_time=hodgepodge.time.to_datetime(min_connector_run_start_time),
        max_connector_run_start_time=hodgepodge.time.to_datetime(max_connector_run_start_time),
        min_connector_run_end_time=hodgepodge.time.to_datetime(min_connector_run_end_time),
        max_connector_run_end_time=hodgepodge.time.to_datetime(max_connector_run_end_time),
        limit=limit,
    )
    count = sum(1 for _ in rows)
    hodgepodge.click.echo_as_json(count)


@connectors.command()
@click.option('--connector-ids')
@click.option('--connector-names')
@click.option('--connector-run-ids')
@click.option('--min-connector-run-start-time')
@click.option('--max-connector-run-start-time')
@click.option('--min-connector-run-end-time')
@click.option('--max-connector-run-end-time')
@click.option('--limit', type=int)
@click.pass_context
def count_connector_runs(
        ctx: click.Context,
        connector_ids: str,
        connector_names: str,
        connector_run_ids: str,
        min_connector_run_start_time: str,
        max_connector_run_start_time: str,
        min_connector_run_end_time: str,
        max_connector_run_end_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    runs = api.get_connector_runs_by_connector_id(
        connector_ids=hodgepodge.click.str_to_list_of_int(connector_ids),
        connector_names=hodgepodge.click.str_to_list_of_str(connector_names),
        connector_run_ids=hodgepodge.click.str_to_list_of_int(connector_run_ids),
        min_connector_run_start_time=hodgepodge.time.to_datetime(min_connector_run_start_time),
        max_connector_run_start_time=hodgepodge.time.to_datetime(max_connector_run_start_time),
        min_connector_run_end_time=hodgepodge.time.to_datetime(min_connector_run_end_time),
        max_connector_run_end_time=hodgepodge.time.to_datetime(max_connector_run_end_time),
        limit=limit,
    )
    count = sum(1 for _ in runs)
    click.echo(count)



from kenna.api import Kenna

import click

import hodgepodge.network
import hodgepodge.click
import hodgepodge.types
import hodgepodge.time


@click.group()
@click.pass_context
def assets(_):
    pass


@assets.command()
@click.option('--asset-id', type=int, required=True)
@click.pass_context
def get_asset(ctx: click.Context, asset_id: int):
    api = Kenna(**ctx.obj['kenna']['config'])
    row = api.get_asset(asset_id=asset_id)
    if row:
        hodgepodge.click.echo_as_json(row)


@assets.command()
@click.option('--asset-ids')
@click.option('--asset-group-ids')
@click.option('--asset-group-names')
@click.option('--asset-tags')
@click.option('--asset-hostnames')
@click.option('--asset-ip-addresses')
@click.option('--asset-mac-addresses')
@click.option('--min-asset-risk-meter-score', type=int)
@click.option('--max-asset-risk-meter-score', type=int)
@click.option('--min-asset-first-seen-time')
@click.option('--max-asset-first-seen-time')
@click.option('--min-asset-last-seen-time')
@click.option('--max-asset-last-seen-time')
@click.option('--min-asset-last-boot-time')
@click.option('--max-asset-last-boot-time')
@click.option('--limit', type=int)
@click.pass_context
def get_assets(
        ctx: click.Context,
        asset_ids: str,
        asset_group_ids: str,
        asset_group_names: str,
        asset_tags: str,
        asset_hostnames: str,
        asset_ip_addresses: str,
        asset_mac_addresses: str,
        min_asset_risk_meter_score: int,
        max_asset_risk_meter_score: int,
        min_asset_first_seen_time: str,
        max_asset_first_seen_time: str,
        min_asset_last_seen_time: str,
        max_asset_last_seen_time: str,
        min_asset_last_boot_time: str,
        max_asset_last_boot_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    for row in api.iter_assets(
        asset_ids=hodgepodge.click.str_to_list_of_int(asset_ids),
        asset_group_ids=hodgepodge.click.str_to_list_of_int(asset_group_ids),
        asset_group_names=hodgepodge.click.str_to_list_of_str(asset_group_names),
        asset_tags=hodgepodge.click.str_to_list_of_str(asset_tags),
        asset_hostnames=hodgepodge.click.str_to_list_of_str(asset_hostnames),
        asset_ip_addresses=hodgepodge.click.str_to_list_of_str(asset_ip_addresses),
        asset_mac_addresses=hodgepodge.click.str_to_list_of_str(asset_mac_addresses),
        min_asset_risk_meter_score=min_asset_risk_meter_score,
        max_asset_risk_meter_score=max_asset_risk_meter_score,
        min_asset_first_seen_time=hodgepodge.time.to_datetime(min_asset_first_seen_time),
        max_asset_first_seen_time=hodgepodge.time.to_datetime(max_asset_first_seen_time),
        min_asset_last_seen_time=hodgepodge.time.to_datetime(min_asset_last_seen_time),
        max_asset_last_seen_time=hodgepodge.time.to_datetime(max_asset_last_seen_time),
        min_asset_last_boot_time=hodgepodge.time.to_datetime(min_asset_last_boot_time),
        max_asset_last_boot_time=hodgepodge.time.to_datetime(max_asset_last_boot_time),
        limit=limit,
    ):
        hodgepodge.click.echo_as_json(row)


@assets.command()
@click.option('--asset-ids')
@click.option('--asset-group-ids')
@click.option('--asset-group-names')
@click.option('--asset-tags')
@click.option('--asset-hostnames')
@click.option('--asset-ip-addresses')
@click.option('--asset-mac-addresses')
@click.option('--min-asset-risk-meter-score', type=int)
@click.option('--max-asset-risk-meter-score', type=int)
@click.option('--min-asset-first-seen-time')
@click.option('--max-asset-first-seen-time')
@click.option('--min-asset-last-seen-time')
@click.option('--max-asset-last-seen-time')
@click.option('--min-asset-last-boot-time')
@click.option('--max-asset-last-boot-time')
@click.option('--limit', type=int)
@click.pass_context
def count_assets(
        ctx: click.Context,
        asset_ids: str,
        asset_group_ids: str,
        asset_group_names: str,
        asset_tags: str,
        asset_hostnames: str,
        asset_ip_addresses: str,
        asset_mac_addresses: str,
        min_asset_risk_meter_score: int,
        max_asset_risk_meter_score: int,
        min_asset_first_seen_time: str,
        max_asset_first_seen_time: str,
        min_asset_last_seen_time: str,
        max_asset_last_seen_time: str,
        min_asset_last_boot_time: str,
        max_asset_last_boot_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    rows = api.iter_assets(
        asset_ids=hodgepodge.click.str_to_list_of_int(asset_ids),
        asset_group_ids=hodgepodge.click.str_to_list_of_int(asset_group_ids),
        asset_group_names=hodgepodge.click.str_to_list_of_str(asset_group_names),
        asset_tags=hodgepodge.click.str_to_list_of_str(asset_tags),
        asset_hostnames=hodgepodge.click.str_to_list_of_str(asset_hostnames),
        asset_ip_addresses=hodgepodge.click.str_to_list_of_str(asset_ip_addresses),
        asset_mac_addresses=hodgepodge.click.str_to_list_of_str(asset_mac_addresses),
        min_asset_risk_meter_score=min_asset_risk_meter_score,
        max_asset_risk_meter_score=max_asset_risk_meter_score,
        min_asset_first_seen_time=hodgepodge.time.to_datetime(min_asset_first_seen_time),
        max_asset_first_seen_time=hodgepodge.time.to_datetime(max_asset_first_seen_time),
        min_asset_last_seen_time=hodgepodge.time.to_datetime(min_asset_last_seen_time),
        max_asset_last_seen_time=hodgepodge.time.to_datetime(max_asset_last_seen_time),
        min_asset_last_boot_time=hodgepodge.time.to_datetime(min_asset_last_boot_time),
        max_asset_last_boot_time=hodgepodge.time.to_datetime(max_asset_last_boot_time),
        limit=limit,
    )
    count = sum(1 for _ in rows)
    click.echo(count)


@assets.command()
@click.option('--asset-ids')
@click.option('--asset-group-ids')
@click.option('--asset-group-names')
@click.option('--asset-tags')
@click.option('--asset-hostnames')
@click.option('--asset-ip-addresses')
@click.option('--asset-mac-addresses')
@click.option('--min-asset-risk-meter-score', type=int)
@click.option('--max-asset-risk-meter-score', type=int)
@click.option('--min-asset-first-seen-time')
@click.option('--max-asset-first-seen-time')
@click.option('--min-asset-last-seen-time')
@click.option('--max-asset-last-seen-time')
@click.option('--min-asset-last-boot-time')
@click.option('--max-asset-last-boot-time')
@click.option('--limit', type=int)
@click.pass_context
def get_asset_hostnames(
        ctx: click.Context,
        asset_ids: str,
        asset_group_ids: str,
        asset_group_names: str,
        asset_tags: str,
        asset_hostnames: str,
        asset_ip_addresses: str,
        asset_mac_addresses: str,
        min_asset_risk_meter_score: int,
        max_asset_risk_meter_score: int,
        min_asset_first_seen_time: str,
        max_asset_first_seen_time: str,
        min_asset_last_seen_time: str,
        max_asset_last_seen_time: str,
        min_asset_last_boot_time: str,
        max_asset_last_boot_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    rows = api.iter_assets(
        asset_ids=hodgepodge.click.str_to_list_of_int(asset_ids),
        asset_group_ids=hodgepodge.click.str_to_list_of_int(asset_group_ids),
        asset_group_names=hodgepodge.click.str_to_list_of_str(asset_group_names),
        asset_tags=hodgepodge.click.str_to_list_of_str(asset_tags),
        asset_hostnames=hodgepodge.click.str_to_list_of_str(asset_hostnames),
        asset_ip_addresses=hodgepodge.click.str_to_list_of_str(asset_ip_addresses),
        asset_mac_addresses=hodgepodge.click.str_to_list_of_str(asset_mac_addresses),
        min_asset_risk_meter_score=min_asset_risk_meter_score,
        max_asset_risk_meter_score=max_asset_risk_meter_score,
        min_asset_first_seen_time=hodgepodge.time.to_datetime(min_asset_first_seen_time),
        max_asset_first_seen_time=hodgepodge.time.to_datetime(max_asset_first_seen_time),
        min_asset_last_seen_time=hodgepodge.time.to_datetime(min_asset_last_seen_time),
        max_asset_last_seen_time=hodgepodge.time.to_datetime(max_asset_last_seen_time),
        min_asset_last_boot_time=hodgepodge.time.to_datetime(min_asset_last_boot_time),
        max_asset_last_boot_time=hodgepodge.time.to_datetime(max_asset_last_boot_time),
        limit=limit,
    )
    hostnames = set(map(str.lower, filter(bool, map(lambda a: a['hostname'], rows))))
    hodgepodge.click.echo_as_json(sorted(hostnames))


@assets.command()
@click.option('--asset-ids')
@click.option('--asset-group-ids')
@click.option('--asset-group-names')
@click.option('--asset-tags')
@click.option('--asset-hostnames')
@click.option('--asset-ip-addresses')
@click.option('--asset-mac-addresses')
@click.option('--min-asset-risk-meter-score', type=int)
@click.option('--max-asset-risk-meter-score', type=int)
@click.option('--min-asset-first-seen-time')
@click.option('--max-asset-first-seen-time')
@click.option('--min-asset-last-seen-time')
@click.option('--max-asset-last-seen-time')
@click.option('--min-asset-last-boot-time')
@click.option('--max-asset-last-boot-time')
@click.option('--limit', type=int)
@click.pass_context
def get_asset_ip_addresses(
        ctx: click.Context,
        asset_ids: str,
        asset_group_ids: str,
        asset_group_names: str,
        asset_tags: str,
        asset_hostnames: str,
        asset_ip_addresses: str,
        asset_mac_addresses: str,
        min_asset_risk_meter_score: int,
        max_asset_risk_meter_score: int,
        min_asset_first_seen_time: str,
        max_asset_first_seen_time: str,
        min_asset_last_seen_time: str,
        max_asset_last_seen_time: str,
        min_asset_last_boot_time: str,
        max_asset_last_boot_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    rows = api.iter_assets(
        asset_ids=hodgepodge.click.str_to_list_of_int(asset_ids),
        asset_group_ids=hodgepodge.click.str_to_list_of_int(asset_group_ids),
        asset_group_names=hodgepodge.click.str_to_list_of_str(asset_group_names),
        asset_tags=hodgepodge.click.str_to_list_of_str(asset_tags),
        asset_hostnames=hodgepodge.click.str_to_list_of_str(asset_hostnames),
        asset_ip_addresses=hodgepodge.click.str_to_list_of_str(asset_ip_addresses),
        asset_mac_addresses=hodgepodge.click.str_to_list_of_str(asset_mac_addresses),
        min_asset_risk_meter_score=min_asset_risk_meter_score,
        max_asset_risk_meter_score=max_asset_risk_meter_score,
        min_asset_first_seen_time=hodgepodge.time.to_datetime(min_asset_first_seen_time),
        max_asset_first_seen_time=hodgepodge.time.to_datetime(max_asset_first_seen_time),
        min_asset_last_seen_time=hodgepodge.time.to_datetime(min_asset_last_seen_time),
        max_asset_last_seen_time=hodgepodge.time.to_datetime(max_asset_last_seen_time),
        min_asset_last_boot_time=hodgepodge.time.to_datetime(min_asset_last_boot_time),
        max_asset_last_boot_time=hodgepodge.time.to_datetime(max_asset_last_boot_time),
        limit=limit,
    )
    ips = set(filter(bool, map(lambda a: a['ip_address'] or a['ipv6'], rows)))
    hodgepodge.click.echo_as_json(sorted(ips))


@assets.command()
@click.option('--asset-ids')
@click.option('--asset-group-ids')
@click.option('--asset-group-names')
@click.option('--asset-tags')
@click.option('--asset-hostnames')
@click.option('--asset-ip-addresses')
@click.option('--asset-mac-addresses')
@click.option('--min-asset-risk-meter-score', type=int)
@click.option('--max-asset-risk-meter-score', type=int)
@click.option('--min-asset-first-seen-time')
@click.option('--max-asset-first-seen-time')
@click.option('--min-asset-last-seen-time')
@click.option('--max-asset-last-seen-time')
@click.option('--min-asset-last-boot-time')
@click.option('--max-asset-last-boot-time')
@click.option('--limit', type=int)
@click.pass_context
def get_asset_mac_addresses(
        ctx: click.Context,
        asset_ids: str,
        asset_group_ids: str,
        asset_group_names: str,
        asset_tags: str,
        asset_hostnames: str,
        asset_ip_addresses: str,
        asset_mac_addresses: str,
        min_asset_risk_meter_score: int,
        max_asset_risk_meter_score: int,
        min_asset_first_seen_time: str,
        max_asset_first_seen_time: str,
        min_asset_last_seen_time: str,
        max_asset_last_seen_time: str,
        min_asset_last_boot_time: str,
        max_asset_last_boot_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    rows = api.iter_assets(
        asset_ids=hodgepodge.click.str_to_list_of_int(asset_ids),
        asset_group_ids=hodgepodge.click.str_to_list_of_int(asset_group_ids),
        asset_group_names=hodgepodge.click.str_to_list_of_str(asset_group_names),
        asset_tags=hodgepodge.click.str_to_list_of_str(asset_tags),
        asset_hostnames=hodgepodge.click.str_to_list_of_str(asset_hostnames),
        asset_ip_addresses=hodgepodge.click.str_to_list_of_str(asset_ip_addresses),
        asset_mac_addresses=hodgepodge.click.str_to_list_of_str(asset_mac_addresses),
        min_asset_risk_meter_score=min_asset_risk_meter_score,
        max_asset_risk_meter_score=max_asset_risk_meter_score,
        min_asset_first_seen_time=hodgepodge.time.to_datetime(min_asset_first_seen_time),
        max_asset_first_seen_time=hodgepodge.time.to_datetime(max_asset_first_seen_time),
        min_asset_last_seen_time=hodgepodge.time.to_datetime(min_asset_last_seen_time),
        max_asset_last_seen_time=hodgepodge.time.to_datetime(max_asset_last_seen_time),
        min_asset_last_boot_time=hodgepodge.time.to_datetime(min_asset_last_boot_time),
        max_asset_last_boot_time=hodgepodge.time.to_datetime(max_asset_last_boot_time),
        limit=limit,
    )
    macs = set(map(hodgepodge.network.parse_mac_address, filter(bool, map(lambda row: row['mac_address'], rows))))
    hodgepodge.click.echo_as_json(sorted(macs))



import click

from kenna import DEFAULT_API_KEY
from kenna.cli.command_groups.applications import applications
from kenna.cli.command_groups.assets import assets
from kenna.cli.command_groups.asset_groups import asset_groups
from kenna.cli.command_groups.connectors import connectors
from kenna.cli.command_groups.dashboard_groups import dashboard_groups
from kenna.cli.command_groups.fixes import fixes
from kenna.cli.command_groups.users import users
from kenna.cli.command_groups.roles import roles
from kenna.cli.command_groups.vulnerabilities import vulnerabilities
from kenna.region import DEFAULT_REGION


@click.group()
@click.option('--api-key', default=DEFAULT_API_KEY)
@click.option('--region', default=DEFAULT_REGION)
@click.pass_context
def cli(ctx, api_key, region):
    ctx.ensure_object(dict)
    ctx.obj['kenna'] = {
        'config': {
            'api_key': api_key,
            'region': region,
        }
    }


COMMAND_GROUPS = [
    applications,
    assets,
    asset_groups,
    connectors,
    dashboard_groups,
    fixes,
    users,
    roles,
    vulnerabilities,
]
for command_group in COMMAND_GROUPS:
    cli.add_command(command_group)



from kenna.api import Kenna

import click

import hodgepodge.click
import hodgepodge.types
import hodgepodge.time


@click.group()
@click.pass_context
def roles(_):
    pass


@roles.command()
@click.option('--role-id', type=int, required=True)
@click.pass_context
def get_role(ctx: click.Context, role_id: int):
    api = Kenna(**ctx.obj['kenna']['config'])
    row = api.get_role(role_id=role_id)
    if row:
        hodgepodge.click.echo_as_json(row)


@roles.command()
@click.option('--role-ids')
@click.option('--role-names')
@click.option('--role-types')
@click.option('--role-access-levels')
@click.option('--role-custom-permissions')
@click.option('--user-ids')
@click.option('--user-email-addresses')
@click.option('--user-names')
@click.option('--min-role-create-time')
@click.option('--max-role-create-time')
@click.option('--min-role-last-update-time')
@click.option('--max-role-last-update-time')
@click.option('--limit', type=int)
@click.pass_context
def get_roles(
        ctx: click.Context,
        role_ids: str,
        role_names: str,
        role_types: str,
        role_access_levels: str,
        role_custom_permissions: str,
        user_ids: str,
        user_email_addresses: str,
        user_names: str,
        min_role_create_time: str,
        max_role_create_time: str,
        min_role_last_update_time: str,
        max_role_last_update_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    for row in api.iter_roles(
        role_ids=hodgepodge.click.str_to_list_of_int(role_ids),
        role_names=hodgepodge.click.str_to_list_of_str(role_names),
        role_types=hodgepodge.click.str_to_list_of_str(role_types),
        role_access_levels=hodgepodge.click.str_to_list_of_str(role_access_levels),
        role_custom_permissions=hodgepodge.click.str_to_list_of_str(role_custom_permissions),
        user_ids=hodgepodge.click.str_to_list_of_int(user_ids),
        user_email_addresses=hodgepodge.click.str_to_list_of_str(user_email_addresses),
        user_names=hodgepodge.click.str_to_list_of_str(user_names),
        min_role_create_time=hodgepodge.time.to_datetime(min_role_create_time),
        max_role_create_time=hodgepodge.time.to_datetime(max_role_create_time),
        min_role_last_update_time=hodgepodge.time.to_datetime(min_role_last_update_time),
        max_role_last_update_time=hodgepodge.time.to_datetime(max_role_last_update_time),
        limit=limit,
    ):
        hodgepodge.click.echo_as_json(row)


@roles.command()
@click.option('--role-ids')
@click.option('--role-names')
@click.option('--role-types')
@click.option('--role-access-levels')
@click.option('--role-custom-permissions')
@click.option('--user-ids')
@click.option('--user-email-addresses')
@click.option('--user-names')
@click.option('--min-role-create-time')
@click.option('--max-role-create-time')
@click.option('--min-role-last-update-time')
@click.option('--max-role-last-update-time')
@click.option('--limit', type=int)
@click.pass_context
def count_roles(
        ctx: click.Context,
        role_ids: str,
        role_names: str,
        role_types: str,
        role_access_levels: str,
        role_custom_permissions: str,
        user_ids: str,
        user_email_addresses: str,
        user_names: str,
        min_role_create_time: str,
        max_role_create_time: str,
        min_role_last_update_time: str,
        max_role_last_update_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    rows = api.iter_roles(
        role_ids=hodgepodge.click.str_to_list_of_int(role_ids),
        role_names=hodgepodge.click.str_to_list_of_str(role_names),
        role_types=hodgepodge.click.str_to_list_of_str(role_types),
        role_access_levels=hodgepodge.click.str_to_list_of_str(role_access_levels),
        role_custom_permissions=hodgepodge.click.str_to_list_of_str(role_custom_permissions),
        user_ids=hodgepodge.click.str_to_list_of_int(user_ids),
        user_email_addresses=hodgepodge.click.str_to_list_of_str(user_email_addresses),
        user_names=hodgepodge.click.str_to_list_of_str(user_names),
        min_role_create_time=hodgepodge.time.to_datetime(min_role_create_time),
        max_role_create_time=hodgepodge.time.to_datetime(max_role_create_time),
        min_role_last_update_time=hodgepodge.time.to_datetime(min_role_last_update_time),
        max_role_last_update_time=hodgepodge.time.to_datetime(max_role_last_update_time),
        limit=limit,
    )
    count = sum(1 for _ in rows)
    click.echo(count)



from typing import Union, Iterable


def parse_variable_names(keys: Union[str, Iterable[str]]):
    keys = keys if isinstance(keys, str) else ','.join(keys)
    for old, new in (
        ('-', '_'),
        (' ', ''),
    ):
        keys = keys.replace(old, new)
    return sorted([key for key in keys.split(',') if key])



from kenna.region import DEFAULT_REGION

import os

DEFAULT_API_KEY = os.environ.get("KENNA_API_KEY")

#: Pagination settings.
PAGE_SIZE_FOR_VULNERABILITIES = 5000
PAGE_SIZE_FOR_ASSETS = 500
PAGE_SIZE_FOR_ASSET_GROUPS = 100
PAGE_SIZE_FOR_FIXES = 100



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kenna', 'kenna.cli', 'kenna.cli.command_groups']

package_data = \
{'': ['*']}

install_requires = \
['DAWG>=0.8.0,<0.9.0',
 'altair>=4.1.0,<5.0.0',
 'antlr4-python3-runtime>=4.9.2,<5.0.0',
 'certifi>=2021.5.30,<2022.0.0',
 'click>=7.0,<8.0',
 'dacite>=1.6.0,<2.0.0',
 'hodgepodge>=1.0.3,<2.0.0',
 'pandas>=1.3.3,<2.0.0',
 'responses>=0.14.0,<0.15.0']

entry_points = \
{'console_scripts': ['kenna = kenna.cli:cli']}

setup_kwargs = {
    'name': 'kenna',
    'version': '0.3.2',
    'description': '',
    'long_description': '# kenna  ![PyPI - Python Version](https://img.shields.io/pypi/pyversions/kenna) ![PyPI - Wheel](https://img.shields.io/pypi/wheel/kenna) ![PyPI - Downloads](https://img.shields.io/pypi/dm/kenna)\n\n![Kenna](https://raw.githubusercontent.com/whitfieldsdad/images/main/kenna-hero.png)\n\n`kenna` is an [API](https://apidocs.kennasecurity.com/reference) client for [Kenna Security](https://www.kennasecurity.com/) that allows you to work with applications, assets, connectors, dashboard groups, users, roles, vulnerabilities, and fixes.\n\n## FAQ\n\n### What\'s included?\n\n- An API client that allows you to lookup and count applications, assets, asset groups, connectors, connector runs, dashboard groups, users, roles, fixes, and vulnerabilities;\n- A robust command-line interface to go along with it\n\n## Installation\n\nTo install `kenna` using `pip`:\n\n```shell\n$ pip install kenna\n```\n\nTo install `kenna` from source (requires [`poetry`](https://github.com/python-poetry/poetry)):\n\n```shell\n$ git clone git@github.com:whitfieldsdad/python-kenna.git\n$ cd python-kenna\n$ make install\n```\n\nTo install `kenna` from source using `setup.py` (i.e. if you\'re not using `poetry`):\n\n```shell\n$ git clone git@github.com:whitfieldsdad/python-kenna.git\n$ cd python-kenna\n$ python3 setup.py install\n```\n \nFinally, export an environment variable named `KENNA_API_KEY` to your shell (i.e. add it to your `~/.bashrc` or user profile).\n\n## Important environment variables\n\n- `KENNA_API_KEY`: your API key for accessing Kenna (required)\n- `ENABLE_INTEGRATION_TESTS`: `true` or `false` (optional)\n\n## Testing\n\nYou can run the unit tests and integration tests for this package as follows:\n\n```shell\n$ make test\n```\n\nIf the value of the `$ENABLE_INTEGRATION_TESTS` environment variable is set to `true`, both the unit tests and integration tests will be executed:\n\n```shell\n$ ENABLE_INTEGRATION_TESTS=true make test\n```\n\nIf you have separate tenants for development and production, you can select which tenant you\'d like to work with as follows:\n\n```shell\n$ ENABLE_INTEGRATION_TESTS=true\n$ KENNA_API_KEY=${KENNA_DEVELOPMENT_API_KEY} make test\n$ KENNA_API_KEY=${KENNA_PRODUCTION_API_KEY} make test\n```\n\n## Tutorials\n\n### Command-line interface\n\nThe command-line interface for `kenna` allows you to search for, list, and count [applications](https://apidocs.kennasecurity.com/reference#list-applications), [assets](https://apidocs.kennasecurity.com/reference#list-assets), [asset groups](https://apidocs.kennasecurity.com/reference#list-asset-groups), [connectors](https://apidocs.kennasecurity.com/reference#list-connectors), [connector runs](https://apidocs.kennasecurity.com/reference#list-connector-runs), [dashboard groups](https://apidocs.kennasecurity.com/reference#list-dashboard-groups), [fixes](https://apidocs.kennasecurity.com/reference#list-fixes), [roles](https://apidocs.kennasecurity.com/reference#list-roles), [users](https://apidocs.kennasecurity.com/reference#list-users), and [vulnerabilities](https://apidocs.kennasecurity.com/reference#list-vulnerabilities).\n\nAfter installing `kenna` you can access the command-line interface as follows:\n\n```shell\n$ poetry run kenna\n```\n\nIf you\'re not using `poetry` (you should use it), you can access the command line as follows:\n\n```shell\n$ python3 -m kenna.cli\n```\n\n#### General\n\nThe following general options are available within the command-line interface:\n\n```shell\n$ poetry run kenna\nUsage: kenna [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --api-key TEXT\n  --region TEXT\n  --help          Show this message and exit.\n\nCommands:\n  applications\n  asset-groups\n  assets\n  connector-runs\n  connectors\n  dashboard-groups\n  fixes\n  roles\n  users\n  vulnerabilities\n````\n\n#### Applications\n\nThe following options are available when working with applications:\n\n```shell\n$ poetry run kenna applications --help\nUsage: kenna applications [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  get-applications\n  count-applications\n```\n\n##### List applications\n\nThe following options are available when listing applications:\n\n```shell\n$ poetry run kenna applications get-applications --help\nUsage: kenna applications get-applications [OPTIONS]\n\nOptions:\n  --application-ids TEXT\n  --application-names TEXT\n  --application-owners TEXT\n  --application-teams TEXT\n  --application-business-units TEXT\n  --min-application-risk-meter-score INTEGER\n  --max-application-risk-meter-score INTEGER\n  --min-asset-count INTEGER\n  --max-asset-count INTEGER\n  --min-vulnerability-count INTEGER\n  --max-vulnerability-count INTEGER\n  --limit INTEGER\n  --help                          Show this message and exit.\n```\n\n#### Assets\n\nThe following options are available when working with assets:\n\n```shell\n$ poetry run kenna assets --help\nUsage: kenna assets [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  get-assets\n  count-assets\n```\n\n##### List assets\n\nThe following options are available when listing assets:\n\n```shell\n$ poetry run kenna assets get-assets --help\nUsage: kenna assets get-assets [OPTIONS]\n\nOptions:\n  --asset-ids TEXT\n  --asset-group-ids TEXT\n  --asset-group-names TEXT\n  --asset-tags TEXT\n  --asset-hostnames TEXT\n  --asset-ip-addresses TEXT\n  --asset-mac-addresses TEXT\n  --min-asset-risk-meter-score INTEGER\n  --max-asset-risk-meter-score INTEGER\n  --min-asset-first-seen-time TEXT\n  --max-asset-first-seen-time TEXT\n  --min-asset-last-seen-time TEXT\n  --max-asset-last-seen-time TEXT\n  --min-asset-last-boot-time TEXT\n  --max-asset-last-boot-time TEXT\n  --limit INTEGER\n  --help                          Show this message and exit.\n```\n\n#### Asset groups\n\nThe following options are available when working with asset groups:\n\n```shell\n$ poetry run kenna asset-groups --help\nUsage: kenna asset-groups [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  get-asset-groups\n  count-asset-groups\n```\n\n##### List asset groups\n\nThe following options are available when listing asset groups:\n\n```shell\n$ poetry run kenna asset-groups get-asset-groups --help\nUsage: kenna asset-groups get-asset-groups [OPTIONS]\n\nOptions:\n  --asset-group-ids TEXT\n  --asset-group-names TEXT\n  --asset-ids TEXT\n  --asset-tags TEXT\n  --asset-hostnames TEXT\n  --asset-ip-addresses TEXT\n  --asset-mac-addresses TEXT\n  --min-asset-group-create-time TEXT\n  --max-asset-group-create-time TEXT\n  --min-asset-group-last-update-time TEXT\n  --max-asset-group-last-update-time TEXT\n  --limit INTEGER\n  --help                          Show this message and exit.\n```\n\nFor example, let\'s list the ID, name, asset count, vulnerability count, and fix count associated with each asset group that has a name containing the string "Europe":\n\n```shell\n$ poetry run kenna asset-groups get-asset-groups --asset-group-names "*Europe*" | jq \'{"id": .id, "name": .name, "asset_count": .asset_count, "vulnerability_count": .vulnerability_count, "fix_count": .fix_count}\' | jq --slurp\n[\n  {\n    "id": 123456,\n    "name": "All European assets",\n    "asset_count": 12345,\n    "vulnerability_count": 1234567,\n    "fix_count": 12345\n  }\n]\n```\n\n#### Connectors\n\nThe following options are available when working with connectors:\n\n```shell\n$ poetry run kenna connectors --help\nUsage: kenna connectors [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  get-connectors\n  count-connectors\n```\n\n##### List connectors\n\nThe following options are available when listing connectors:\n\n```shell\n$ poetry run kenna connectors get-connectors --help\nUsage: kenna connectors get-connectors [OPTIONS]\n\nOptions:\n  --connector-ids TEXT\n  --connector-names TEXT\n  --min-connector-run-start-time TEXT\n  --max-connector-run-start-time TEXT\n  --min-connector-run-end-time TEXT\n  --max-connector-run-end-time TEXT\n  --limit INTEGER\n  --help                          Show this message and exit.\n```\n\n#### Connector runs\n\nThe following options are available when working with connector runs:\n\n```shell\n$ poetry run kenna connector-runs --help\n\nUsage: kenna connector-runs [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --connector-ids TEXT\n  --connector-names TEXT\n  --connector-run-ids TEXT\n  --min-connector-run-start-time TEXT\n  --max-connector-run-start-time TEXT\n  --min-connector-run-end-time TEXT\n  --max-connector-run-end-time TEXT\n  --help                          Show this message and exit.\n\nCommands:\n  get-connector-runs\n  count-connector-runs\n```\n\n##### List connector runs\n\nThe following options are available when listing connector runs:\n\n```shell\n$ poetry run kenna connector-runs get-connector-runs --help\nUsage: kenna connector-runs get-connector-runs [OPTIONS]\n\nOptions:\n  --connector-ids TEXT\n  --connector-names TEXT\n  --connector-run-ids TEXT\n  --min-connector-run-start-time TEXT\n  --max-connector-run-start-time TEXT\n  --min-connector-run-end-time TEXT\n  --max-connector-run-end-time TEXT\n  --limit INTEGER\n  --help                          Show this message and exit.\n```\n\n#### Dashboard groups\n\nThe following options are available when working with dashboard groups:\n\n```shell\n$ poetry run kenna dashboard-groups --help\nUsage: kenna dashboard-groups [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  get-dashboard-groups\n  count-dashboard-groups\n```\n\n##### List dashboard groups\n\nThe following options are available when listing dashboard groups:\n\n```shell\n$ poetry run kenna dashboard-groups get-dashboard-groups --help\nUsage: kenna dashboard-groups get-dashboard-groups [OPTIONS]\n\nOptions:\n  --dashboard-group-ids TEXT\n  --dashboard-group-names TEXT\n  --role-ids TEXT\n  --role-names TEXT\n  --min-dashboard-group-create-time TEXT\n  --max-dashboard-group-create-time TEXT\n  --min-dashboard-group-last-update-time TEXT\n  --max-dashboard-group-last-update-time TEXT\n  --limit INTEGER\n  --help                          Show this message and exit.\n```\n\n#### Users\n\nThe following options are available when working with users:\n\n```shell\n$ poetry run kenna users --help\nUsage: kenna users [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  get-users\n  count-users\n```\n\n##### List users\n\nThe following options are available when listing users:\n\n```shell\n$ poetry run kenna users get-users --help\npoetry run kenna users get-users --help\nUsage: kenna users get-users [OPTIONS]\n\nOptions:\n  --user-ids TEXT\n  --user-names TEXT\n  --user-email-addresses TEXT\n  --min-user-create-time TEXT\n  --max-user-create-time TEXT\n  --min-user-last-sign-in-time TEXT\n  --max-user-last-sign-in-time TEXT\n  --min-user-last-update-time TEXT\n  --max-user-last-update-time TEXT\n  --limit INTEGER\n  --help                          Show this message and exit.\n```\n\n#### Roles\n\nThe following options are available when working with roles:\n\n```shell\n$ poetry run kenna roles --help\nUsage: kenna roles [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  get-roles\n  count-roles\n```\n\n##### List roles\n\nThe following options are available when listing roles:\n\n```shell\n$ poetry run kenna roles get-roles --help\nUsage: kenna roles get-roles [OPTIONS]\n\nOptions:\n  --role-ids TEXT\n  --role-names TEXT\n  --role-types TEXT\n  --role-access-levels TEXT\n  --role-custom-permissions TEXT\n  --user-ids TEXT\n  --user-email-addresses TEXT\n  --user-names TEXT\n  --min-role-create-time TEXT\n  --max-role-create-time TEXT\n  --min-role-last-update-time TEXT\n  --max-role-last-update-time TEXT\n  --limit INTEGER\n  --help                          Show this message and exit.\n```\n\n#### Vulnerabilities\n\nThe following options are available when working with vulnerabilities:\n\n```shell\n$ poetry run kenna vulnerabilities --help\nUsage: kenna vulnerabilities [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  get-vulnerabilities\n  count-vulnerabilities\n```\n\n##### List vulnerabilities\n\nThe following options are available when listing vulnerabilities:\n\n```shell\n$ poetry run kenna vulnerabilities get-vulnerabilities --help\nUsage: kenna vulnerabilities get-vulnerabilities [OPTIONS]\n\nOptions:\n  --vulnerability-ids TEXT\n  --cve-ids TEXT\n  --fix-ids TEXT\n  --fix-names TEXT\n  --fix-vendors TEXT\n  --asset-ids TEXT\n  --asset-hostnames TEXT\n  --asset-ip-addresses TEXT\n  --asset-mac-addresses TEXT\n  --asset-group-ids TEXT\n  --asset-group-names TEXT\n  --asset-tags TEXT\n  --min-vulnerability-risk-meter-score INTEGER\n  --max-vulnerability-risk-meter-score INTEGER\n  --min-vulnerability-first-seen-time TEXT\n  --max-vulnerability-first-seen-time TEXT\n  --min-vulnerability-last-seen-time TEXT\n  --max-vulnerability-last-seen-time TEXT\n  --min-cve-publish-time TEXT\n  --max-cve-publish-time TEXT\n  --min-patch-publish-time TEXT\n  --max-patch-publish-time TEXT\n  --min-patch-due-date TEXT\n  --max-patch-due-date TEXT\n  --limit INTEGER\n  --help                          Show this message and exit.\n```\n\n#### Fixes\n\nThe following options are available when working with fixes:\n\n```shell\n$ poetry run kenna fixes --help\nUsage: kenna fixes [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  get-fixes\n  count-fixes\n```\n\n##### List fixes\n\nThe following options are available when listing fixes:\n\n```shell\n$ poetry run kenna fixes get-fixes --help\nUsage: kenna fixes get-fixes [OPTIONS]\n\nOptions:\n  --fix-ids TEXT\n  --fix-names TEXT\n  --fix-vendors TEXT\n  --cve-ids TEXT\n  --asset-ids TEXT\n  --asset-hostnames TEXT\n  --asset-ip-addresses TEXT\n  --asset-mac-addresses TEXT\n  --asset-ip-addresses TEXT\n  --asset-mac-addresses TEXT\n  --asset-group-ids TEXT\n  --asset-group-names TEXT\n  --asset-tags TEXT\n  --min-fix-create-time TEXT\n  --max-fix-create-time TEXT\n  --min-fix-last-update-time TEXT\n  --max-fix-last-update-time TEXT\n  --limit INTEGER\n  --help                          Show this message and exit.\n```\n',
    'author': 'Tyler Fisher',
    'author_email': 'tylerfisher@tylerfisher.ca',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4',
}


setup(**setup_kwargs)



from kenna.error import MissingCredentialsError, InvalidCredentialsError

import os

DEFAULT_API_KEY = os.getenv('KENNA_API_KEY')

API_KEY_LENGTH = 64


def validate_api_key(key: str) -> str:
    if not key:
        raise MissingCredentialsError("An API key is required")

    if len(key) != API_KEY_LENGTH:
        raise InvalidCredentialsError("API key must be {} characters in length - not {}".format(API_KEY_LENGTH, len(key)))

    return key



class _CredentialError(ValueError):
    pass


class MissingCredentialsError(_CredentialError):
    pass


class InvalidCredentialsError(_CredentialError):
    pass


class InvalidRegionError(ValueError):
    pass



from kenna.api import Kenna

import click

import hodgepodge.click
import hodgepodge.types
import hodgepodge.time


@click.group()
@click.pass_context
def users(_):
    pass


@users.command()
@click.option('--user-id', type=int, required=True)
@click.pass_context
def get_user(ctx: click.Context, user_id: int):
    api = Kenna(**ctx.obj['kenna']['config'])
    row = api.get_user(user_id=user_id)
    if row:
        hodgepodge.click.echo_as_json(row)


@users.command()
@click.option('--user-ids')
@click.option('--user-names')
@click.option('--user-email-addresses')
@click.option('--min-user-create-time')
@click.option('--max-user-create-time')
@click.option('--min-user-last-sign-in-time')
@click.option('--max-user-last-sign-in-time')
@click.option('--min-user-last-update-time')
@click.option('--max-user-last-update-time')
@click.option('--limit', type=int)
@click.pass_context
def get_users(
        ctx: click.Context,
        user_ids: str,
        user_names: str,
        user_email_addresses: str,
        min_user_create_time: str,
        max_user_create_time: str,
        min_user_last_update_time: str,
        max_user_last_update_time: str,
        min_user_last_sign_in_time: str,
        max_user_last_sign_in_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    for row in api.iter_users(
        user_ids=hodgepodge.click.str_to_list_of_int(user_ids),
        user_names=hodgepodge.click.str_to_list_of_str(user_names),
        user_email_addresses=hodgepodge.click.str_to_list_of_str(user_email_addresses),
        min_user_create_time=hodgepodge.time.to_datetime(min_user_create_time),
        max_user_create_time=hodgepodge.time.to_datetime(max_user_create_time),
        min_user_last_update_time=hodgepodge.time.to_datetime(min_user_last_update_time),
        max_user_last_update_time=hodgepodge.time.to_datetime(max_user_last_update_time),
        min_user_last_sign_in_time=hodgepodge.time.to_datetime(min_user_last_sign_in_time),
        max_user_last_sign_in_time=hodgepodge.time.to_datetime(max_user_last_sign_in_time),
        limit=limit,
    ):
        hodgepodge.click.echo_as_json(row)


@users.command()
@click.option('--user-ids')
@click.option('--user-names')
@click.option('--user-email-addresses')
@click.option('--min-user-create-time')
@click.option('--max-user-create-time')
@click.option('--min-user-last-sign-in-time')
@click.option('--max-user-last-sign-in-time')
@click.option('--min-user-last-update-time')
@click.option('--max-user-last-update-time')
@click.option('--limit', type=int)
@click.pass_context
def count_users(
        ctx: click.Context,
        user_ids: str,
        user_names: str,
        user_email_addresses: str,
        min_user_create_time: str,
        max_user_create_time: str,
        min_user_last_update_time: str,
        max_user_last_update_time: str,
        min_user_last_sign_in_time: str,
        max_user_last_sign_in_time: str,
        limit: int):

    api = Kenna(**ctx.obj['kenna']['config'])
    rows = api.iter_users(
        user_ids=hodgepodge.click.str_to_list_of_int(user_ids),
        user_names=hodgepodge.click.str_to_list_of_str(user_names),
        user_email_addresses=hodgepodge.click.str_to_list_of_str(user_email_addresses),
        min_user_create_time=hodgepodge.time.to_datetime(min_user_create_time),
        max_user_create_time=hodgepodge.time.to_datetime(max_user_create_time),
        min_user_last_update_time=hodgepodge.time.to_datetime(min_user_last_update_time),
        max_user_last_update_time=hodgepodge.time.to_datetime(max_user_last_update_time),
        min_user_last_sign_in_time=hodgepodge.time.to_datetime(min_user_last_sign_in_time),
        max_user_last_sign_in_time=hodgepodge.time.to_datetime(max_user_last_sign_in_time),
        limit=limit,
    )
    count = sum(1 for _ in rows)
    click.echo(count)
